"""
 @file
 @brief Unit tests for anonymous metrics delivery
 @author OpenShot Studios, LLC

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 """

import os
import sys
import types
import unittest
from unittest.mock import Mock, patch

from qt_api import QApplication


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from tests.qt_test_app import ensure_app_state, get_or_create_app


class DummySettings:
    def __init__(self, send_metrics=True):
        self.values = {
            "send_metrics": send_metrics,
            "unique_install_id": "test-install-id",
        }

    def get(self, key):
        return self.values.get(key, False)


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])
        self.settings = DummySettings()

    def get_settings(self):
        return self.settings


app, _owns_app = get_or_create_app(DummyApp)
ensure_app_state(app, DummySettings, extra_attrs={"window": types.SimpleNamespace()})

from classes import metrics


class MetricsTests(unittest.TestCase):
    def setUp(self):
        metrics.s = DummySettings(send_metrics=True)
        metrics.metric_queue.clear()
        metrics.metric_worker_active = False

    def tearDown(self):
        metrics.metric_queue.clear()
        metrics.metric_worker_active = False

    def test_send_metric_starts_background_worker_without_posting_inline(self):
        thread_instance = Mock()
        with patch.object(metrics.threading, "Thread", return_value=thread_instance) as thread_cls, \
                patch.object(metrics.http_client, "post_json") as post:
            metrics.send_metric({"name": "screen_view", "params": {}})

        post.assert_not_called()
        thread_cls.assert_called_once()
        thread_instance.start.assert_called_once()
        self.assertTrue(metrics.metric_worker_active)
        self.assertEqual(len(metrics.metric_queue), 1)

    def test_metric_worker_uses_short_timeout_and_clears_queue(self):
        metrics.metric_queue.append({"name": "screen_view", "params": {}})
        metrics.metric_worker_active = True
        response = types.SimpleNamespace(status_code=204)

        with patch.object(metrics.http_client, "post_json", return_value=response) as post, \
                patch.object(metrics.time, "sleep"):
            metrics._send_metric_worker()

        post.assert_called_once()
        self.assertEqual(
            post.call_args.kwargs["timeout"],
            (metrics.METRIC_CONNECT_TIMEOUT, metrics.METRIC_READ_TIMEOUT),
        )
        self.assertFalse(metrics.metric_queue)
        self.assertFalse(metrics.metric_worker_active)

    def test_metric_worker_requeues_failed_payload(self):
        event = {"name": "screen_view", "params": {}}
        metrics.metric_queue.append(event)
        metrics.metric_worker_active = True

        with patch.object(metrics.http_client, "post_json", side_effect=RuntimeError("blocked")), \
                patch.object(metrics.log, "warning"):
            metrics._send_metric_worker()

        self.assertEqual(metrics.metric_queue, [event])
        self.assertFalse(metrics.metric_worker_active)


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Targeted unit tests for Project Files export helpers.
"""

import importlib
import os
import sys
import types
import unittest


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from qt_api import QApplication
from tests.qt_test_app import ensure_app_state, get_or_create_app


class DummySettings:
    pass


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])

    def _tr(self, text):
        return text


class ExportClipsTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app, cls._owns_app = get_or_create_app(DummyApp)
        cls.app = ensure_app_state(app, DummySettings)
        cls.export_clips = importlib.import_module("windows.export_clips")

    def test_image_sequence_detection_matches_printf_pattern(self):
        sequence = types.SimpleNamespace(data={"path": os.path.join("titles", "Title%04d.png")})
        normal_file = types.SimpleNamespace(data={"path": os.path.join("titles", "Title0001.png")})

        self.assertTrue(self.export_clips.isImageSequence(sequence))
        self.assertFalse(self.export_clips.isImageSequence(normal_file))

    def test_image_sequence_export_name_is_single_mp4(self):
        file_obj = types.SimpleNamespace(data={
            "path": os.path.join("titles", "Title%04d.png"),
            "fps": {"num": 25, "den": 1},
            "video_length": 300,
        })

        self.assertEqual(
            self.export_clips.nameOfImageSequenceExport(file_obj),
            "title [0.00 - 12.00].mp4",
        )

    def test_image_sequence_as_clip_sets_full_duration_range(self):
        file_obj = types.SimpleNamespace(data={
            "path": os.path.join("titles", "Title%04d.png"),
            "fps": {"num": 25, "den": 1},
            "video_length": 300,
            "width": 1920,
        })

        clip_obj = self.export_clips.imageSequenceAsClip(file_obj)

        self.assertEqual(clip_obj.data["start"], 0.0)
        self.assertEqual(clip_obj.data["end"], 12.0)
        self.assertEqual(clip_obj.data["path"], os.path.join("titles", "Title%04d.png"))
        self.assertEqual(clip_obj.data["width"], 1920)

    def test_setup_writer_disables_audio_for_silent_image_sequence(self):
        clip_obj = types.SimpleNamespace(data={
            "path": os.path.join("titles", "Title%04d.png"),
            "fps": {"num": 25, "den": 1},
            "pixel_ratio": {"num": 1, "den": 1},
            "width": 1920,
            "height": 1080,
            "has_audio": False,
            "sample_rate": 0,
            "channels": 0,
            "channel_layout": 0,
        })
        writer = types.SimpleNamespace(
            audio_options=None,
            video_options=None,
            prepare_count=0,
            opened=False,
        )
        writer.SetVideoOptions = lambda *args: setattr(writer, "video_options", args)
        writer.SetAudioOptions = lambda *args: setattr(writer, "audio_options", args)
        writer.PrepareStreams = lambda: setattr(writer, "prepare_count", writer.prepare_count + 1)
        writer.Open = lambda: setattr(writer, "opened", True)

        self.export_clips.setupWriter(clip_obj, writer)

        self.assertFalse(writer.audio_options[0])
        self.assertEqual(writer.audio_options[2], 48000)
        self.assertEqual(writer.audio_options[3], 2)
        self.assertEqual(writer.audio_options[4], 3)
        self.assertEqual(writer.prepare_count, 2)
        self.assertTrue(writer.opened)


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Unit tests for local path normalization helpers
"""

import os
import sys
import unittest


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes.path_utils import comparable_local_path, normalized_local_path


class PathUtilsTests(unittest.TestCase):
    def test_windows_drive_paths_are_treated_as_local_paths(self):
        path = r"C:\Projects\example_assets\title\intro.svg"

        self.assertEqual(normalized_local_path(path), os.path.normpath(os.path.abspath(path)))
        self.assertEqual(comparable_local_path(path), os.path.normcase(os.path.normpath(os.path.abspath(path))))

    def test_content_uris_remain_opaque(self):
        path = "content://documents/tree/project.osp"

        self.assertEqual(normalized_local_path(path), path)
        self.assertEqual(comparable_local_path(path), path)



"""Regression tests for location-aware desktop file dialogs."""

import os
import sys
import unittest
from unittest.mock import patch


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

import qt_api


class LocationFileDialogTests(unittest.TestCase):
    def test_portal_theme_uses_qt_dialog(self):
        expected = qt_api.QtWidgets.QFileDialog.DontUseNativeDialog
        with patch.dict(os.environ, {"QT_QPA_PLATFORMTHEME": "xdgdesktopportal"}):
            self.assertEqual(qt_api.location_file_dialog_options(), expected)

    def test_other_themes_keep_native_dialog(self):
        with patch.dict(os.environ, {"QT_QPA_PLATFORMTHEME": "gtk3"}):
            self.assertIsNone(qt_api.location_file_dialog_options())

    def test_open_dialog_preserves_directory_with_portal_theme(self):
        selected = [qt_api.QtCore.QUrl.fromLocalFile("/media/recent/video.mp4")]
        callback_results = []

        with patch.dict(os.environ, {"QT_QPA_PLATFORMTHEME": "xdgdesktopportal"}), \
                patch.object(qt_api, "_is_android_runtime", return_value=False), \
                patch.object(qt_api.QtWidgets.QFileDialog, "getOpenFileUrls",
                             return_value=(selected, "")) as get_urls:
            qt_api.show_open_file_dialog(
                None, "Import", "/media/recent", "", callback_results.extend)

        self.assertEqual(callback_results, selected)
        self.assertEqual(get_urls.call_args.args[2].toLocalFile(), "/media/recent")
        self.assertEqual(
            get_urls.call_args.kwargs["options"],
            qt_api.QtWidgets.QFileDialog.DontUseNativeDialog,
        )

    def test_save_dialog_preserves_directory_with_portal_theme(self):
        callback_results = []

        with patch.dict(os.environ, {"QT_QPA_PLATFORMTHEME": "xdgdesktopportal"}), \
                patch.object(qt_api, "_is_android_runtime", return_value=False), \
                patch.object(qt_api.QtWidgets.QFileDialog, "getSaveFileName",
                             return_value=("/projects/recent/movie.osp", "")) as get_name:
            qt_api.show_save_file_dialog(
                None, "Save", "movie.osp", "*/*", callback_results.append,
                directory="/projects/recent",
            )

        self.assertEqual(callback_results, ["/projects/recent/movie.osp"])
        self.assertEqual(get_name.call_args.args[2], "/projects/recent/movie.osp")
        self.assertEqual(
            get_name.call_args.kwargs["options"],
            qt_api.QtWidgets.QFileDialog.DontUseNativeDialog,
        )


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief This file contains unit tests for selected main window flows
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import importlib
import json
import os
import sys
import tempfile
import threading
import types
import unittest
import zipfile
from contextlib import ExitStack
from datetime import datetime, timedelta
from unittest.mock import MagicMock, patch

import openshot

PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from qt_api import QByteArray, QCoreApplication, QEvent, QKeySequence, Qt
from qt_api import QApplication, QDockWidget, QMainWindow, QMenu, QStandardItem, QStandardItemModel

from classes import qt_types
from classes.project_data import ProjectDataStore
from classes.settings import SettingStore
from classes.updates import UpdateManager
from tests.qt_test_app import ensure_app_state as ensure_qt_app_state, get_or_create_app

QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)


class DummySettings:
    actionType = types.SimpleNamespace(LOAD="load")

    def __init__(self):
        self.values = {
            "recent_projects": [],
            "history-limit": 20,
            "recovery-limit": 10,
            "default-profile": "HD 720p 30 fps",
            "default-samplerate": 48000,
            "default-channels": 2,
        }
        self.default_paths = {}

    def get(self, key):
        return self.values.get(key)

    def set(self, key, value):
        self.values[key] = value

    def setDefaultPath(self, action, path):
        self.default_paths[action] = path


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])
        self.settings = DummySettings()
        self.project = None
        self.updates = None
        self.window = None
        self.logger_libopenshot = None

    def get_settings(self):
        return self.settings

    def _tr(self, text):
        return text


def ensure_app_state(app):
    return ensure_qt_app_state(
        app,
        DummySettings,
        project_factory=ProjectDataStore,
        updates_factory=UpdateManager,
        extra_attrs={"window": None, "logger_libopenshot": None},
    )


class SignalRecorder:
    def __init__(self):
        self.calls = []

    def emit(self, *args):
        self.calls.append(args)


class FakeDock:
    def __init__(self, name="", height=100, width=200, area=Qt.TopDockWidgetArea):
        self._name = name
        self._height = height
        self._width = width
        self._area = area
        self._minimum_height = 0
        self._maximum_height = 16777215
        self._minimum_width = 0
        self._maximum_width = 16777215
        self.fixed_heights = []
        self.fixed_widths = []

    def objectName(self):
        return self._name

    def height(self):
        return self._height

    def width(self):
        return self._width

    def minimumHeight(self):
        return self._minimum_height

    def maximumHeight(self):
        return self._maximum_height

    def minimumWidth(self):
        return self._minimum_width

    def maximumWidth(self):
        return self._maximum_width

    def setFixedHeight(self, height):
        self.fixed_heights.append(height)
        self._height = height
        self._minimum_height = height
        self._maximum_height = height

    def setFixedWidth(self, width):
        self.fixed_widths.append(width)
        self._width = width
        self._minimum_width = width
        self._maximum_width = width

    def setMinimumHeight(self, height):
        self._minimum_height = height

    def setMaximumHeight(self, height):
        self._maximum_height = height

    def setMinimumWidth(self, width):
        self._minimum_width = width

    def setMaximumWidth(self, width):
        self._maximum_width = width


class MainWindowTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app, cls._owns_app = get_or_create_app(DummyApp)
        cls.app = ensure_app_state(app)
        metrics = types.ModuleType("classes.metrics")
        metrics.track_metric_session = lambda *args, **kwargs: None
        metrics.track_metric_screen = lambda *args, **kwargs: None
        sys.modules["classes.metrics"] = metrics
        sys.modules.pop("windows.views.timeline", None)
        sys.modules.pop("windows.main_window", None)
        sys.modules.pop("windows.views.properties_tableview", None)
        sys.modules.pop("windows.models.properties_model", None)
        cls.main_window_module = importlib.import_module("windows.main_window")
        cls.properties_tableview_module = importlib.import_module("windows.views.properties_tableview")
        cls.properties_model_module = importlib.import_module("windows.models.properties_model")

    @classmethod
    def tearDownClass(cls):
        if getattr(cls, "_owns_app", False) and cls.app:
            cls.app.quit()

    def test_track_stack_preserves_selected_track_across_renumbering(self):
        module = self.main_window_module
        layers = [
            {"id": "selected", "number": 1},
            {"id": "upper", "number": 1000000},
        ]

        class FakeTrack:
            def __init__(self, layer):
                self.id = layer["id"]
                self.data = layer

            @classmethod
            def get(cls, number=None, id=None):
                for layer in layers:
                    if number is not None and layer["number"] == number:
                        return cls(layer)
                    if id is not None and layer["id"] == id:
                        return cls(layer)
                return None

        helper = types.SimpleNamespace()
        helper._track_numbers = lambda: sorted(layer["number"] for layer in layers)
        helper.ensure_tracks_for_layers = lambda _layers: None

        def create_track_below(number):
            if min(helper._track_numbers()) <= 2:
                layers[0]["number"] = 2000000
                layers[1]["number"] = 3000000
                layers.append({"id": "lower-1", "number": 1000000})
                return 1000000
            new_number = int(number / 2)
            layers.append({"id": "lower-2", "number": new_number})
            return new_number

        helper.create_track_below = create_track_below

        with patch.object(module, "Track", FakeTrack):
            tracks = module.MainWindow.track_stack_from(helper, 1, 3)

        self.assertEqual(tracks, [2000000, 1000000, 500000])
        self.assertNotIn(1, tracks)

    def test_cosmic_play_toggle_uses_hidpi_svg_renderer_for_both_states(self):
        from themes.cosmic.theme import CosmicTheme

        button = types.SimpleNamespace(setIcon=MagicMock())
        icon_size = object()
        action_play = object()
        toolbar = types.SimpleNamespace(
            widgetForAction=lambda action: button if action is action_play else None,
            iconSize=lambda: icon_size,
        )
        theme = CosmicTheme.__new__(CosmicTheme)
        theme.app = types.SimpleNamespace(
            window=types.SimpleNamespace(videoToolbar=toolbar, actionPlay=action_play)
        )
        theme.create_svg_icon = MagicMock(side_effect=["pause-icon", "play-icon"])

        theme.togglePlayIcon(True)
        pause_path, pause_size = theme.create_svg_icon.call_args.args
        self.assertEqual(os.path.basename(pause_path), "tool-media-pause.svg")
        self.assertIs(pause_size, icon_size)
        button.setIcon.assert_called_with("pause-icon")

        theme.togglePlayIcon(False)
        play_path, play_size = theme.create_svg_icon.call_args.args
        self.assertEqual(os.path.basename(play_path), "tool-media-play.svg")
        self.assertIs(play_size, icon_size)
        button.setIcon.assert_called_with("play-icon")

    def setUp(self):
        ensure_app_state(self.app)
        self.app.settings = DummySettings()
        self.app.window = None

    def tearDown(self):
        ensure_app_state(self.app)

    def test_manage_recovery_files_keeps_daily_and_historical_limits(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            recovery_dir = os.path.join(tmpdir, "recovery")
            os.mkdir(recovery_dir)

            with patch.object(self.main_window_module.info, "RECOVERY_PATH", recovery_dir):
                # Freeze the module clock so the retention logic and fixture
                # timestamps use the same notion of "today".
                now = datetime(2026, 3, 21, 12, 0, 0)

                class FixedDateTime(datetime):
                    @classmethod
                    def now(cls, tz=None):
                        if tz is not None:
                            return tz.fromutc(now.replace(tzinfo=tz))
                        return now

                files = [
                    ("100-newest-project.zip", now),
                    ("090-older-today-project.zip", now - timedelta(hours=1)),
                    ("080-yesterday-project.zip", now - timedelta(days=1)),
                    ("070-two-days-project.zip", now - timedelta(days=2)),
                ]
                for name, dt in files:
                    path = os.path.join(recovery_dir, name)
                    with open(path, "w", encoding="utf-8") as handle:
                        handle.write("x")
                    ts = dt.timestamp()
                    os.utime(path, (ts, ts))

                fake_window = types.SimpleNamespace()
                with patch.object(self.main_window_module, "datetime", FixedDateTime):
                    self.main_window_module.MainWindow.manage_recovery_files(fake_window, 1, 1, "project")

                remaining = sorted(os.listdir(recovery_dir))
                self.assertEqual(remaining, ["080-yesterday-project.zip", "100-newest-project.zip"])

    def test_dock_style_scheduler_ignores_signal_payloads(self):
        starts = []
        callbacks = []

        class FakeSignal:
            def connect(self, callback):
                callbacks.append(callback)

        class FakeTimer:
            def __init__(self, parent=None):
                self.timeout = FakeSignal()

            def setSingleShot(self, enabled):
                pass

            def start(self, delay):
                starts.append(delay)

        fake_window = types.SimpleNamespace(
            _apply_scheduled_dock_style_update=lambda: None,
        )

        with patch.object(self.main_window_module, "QTimer", FakeTimer):
            self.main_window_module.MainWindow._schedule_dock_style_update(
                fake_window,
                True,
            )
            self.assertFalse(fake_window._dock_style_theme_changed)
            self.assertEqual(starts[-1], 150)

            self.main_window_module.MainWindow._schedule_dock_style_update(
                fake_window,
                theme_changed=True,
                delay=0,
            )
            self.assertTrue(fake_window._dock_style_theme_changed)
            self.assertEqual(starts[-1], 0)

    def test_dock_top_level_change_marks_interaction_and_restyles_immediately(self):
        calls = []
        fake_window = types.SimpleNamespace(
            _mark_dock_interaction_active=lambda: calls.append("interaction"),
            _schedule_dock_style_update=lambda **kwargs: calls.append(("style", kwargs)),
        )

        self.main_window_module.MainWindow._on_dock_top_level_changed(fake_window, True)

        self.assertEqual(calls, ["interaction", ("style", {"delay": 0})])

    def test_save_settings_uses_serialized_qt_state(self):
        timeline_dock = FakeDock("dockTimeline", height=240, area=Qt.BottomDockWidgetArea)
        video_dock = FakeDock("dockVideo", height=360, width=520, area=Qt.TopDockWidgetArea)
        fake_window = types.SimpleNamespace(
            dockTimeline=timeline_dock,
            dockVideo=video_dock,
            saveState=lambda: QByteArray(b"state"),
            saveGeometry=lambda: QByteArray(b"geometry"),
            getDocks=lambda: [timeline_dock, video_dock],
            dockWidgetArea=lambda dock: dock._area,
        )

        self.main_window_module.MainWindow.save_settings(fake_window)

        saved = qt_types.str_to_bytes(
            self.app.settings.values["window_state_v2"])
        self.assertEqual(bytes(saved), b"state")
        self.assertEqual(self.app.settings.values["video_dock_width"], 520)

    def test_save_settings_prefers_geometry_captured_before_close_dialog(self):
        fake_window = types.SimpleNamespace(
            _pending_close_geometry=QByteArray(b"maximized-geometry"),
            dockTimeline=None,
            dockVideo=None,
            saveState=lambda: QByteArray(b"state"),
            saveGeometry=lambda: QByteArray(b"dialog-altered-geometry"),
            getDocks=lambda: [],
            dockWidgetArea=lambda dock: Qt.NoDockWidgetArea,
        )

        self.main_window_module.MainWindow.save_settings(fake_window)

        saved = qt_types.str_to_bytes(
            self.app.settings.values["window_geometry_v2"])
        self.assertEqual(bytes(saved), b"maximized-geometry")
        self.assertIsNone(fake_window._pending_close_geometry)

    def test_default_settings_include_video_dock_width(self):
        with open(os.path.join(PATH, "settings", "_default.settings"), encoding="utf-8") as fh:
            settings = json.load(fh)

        values = {item.get("setting"): item.get("value") for item in settings}

        self.assertIn("video_dock_width", values)
        self.assertEqual(values["active_builtin_view"], "simple")

    def test_setting_store_invalid_key_warns_without_crashing(self):
        store = SettingStore()
        store._data = [{"setting": "known", "value": 1}]

        store.set("unknown", 2)

        self.assertEqual(store.get("known"), 1)

    def test_project_files_focus_blocks_properties_shortcut_for_keyboard_search(self):
        event = types.SimpleNamespace(
            type=lambda: QEvent.ShortcutOverride,
            modifiers=lambda: Qt.NoModifier,
            key=lambda: Qt.Key_U,
            accept=MagicMock(),
        )
        inactive_view = types.SimpleNamespace(hasFocus=lambda: False)
        window = types.SimpleNamespace(
            focusWidget=lambda: object(),
            _blocks_timeline_shortcuts=lambda _widget: False,
            emojiListView=inactive_view,
            filesView=types.SimpleNamespace(hasFocus=lambda: True),
            transitionsView=inactive_view,
            effectsView=inactive_view,
            getShortcutByName=lambda name: (
                [QKeySequence("U")] if name == "actionProperties" else []
            ),
        )
        self.app.window = window

        handled = self.main_window_module.MainWindow.eventFilter(window, None, event)

        self.assertTrue(handled)
        event.accept.assert_called_once_with()

    def test_startup_restore_uses_qt_state_without_scalar_dock_overrides(self):
        calls = []
        fake_window = types.SimpleNamespace(
            _restored_saved_window=False,
            saved_state=QByteArray(b"state"),
            restoreState=lambda state: calls.append(("state", state)),
            _restore_hidden_docks=lambda names: calls.append(("hidden", names)),
            _restore_saved_video_dock_width=lambda: calls.append(("video",)),
        )

        with patch.object(
                self.main_window_module.QTimer,
                "singleShot",
                lambda delay, callback: calls.append(("timer", delay, callback))):
            self.main_window_module.MainWindow._restore_saved_window_state(
                fake_window)

        self.assertEqual(calls[:2], [
            ("state", QByteArray(b"state")),
            ("hidden", []),
        ])
        self.assertEqual(calls[2][:2], ("timer", 0))

    def test_video_width_correction_resizes_one_horizontal_group(self):
        files = FakeDock("dockFiles", width=700)
        video = FakeDock("dockVideo", width=300)
        calls = []
        fake_window = types.SimpleNamespace(
            dockFiles=files,
            dockVideo=video,
            saved_video_dock_width=450,
            resizeDocks=lambda docks, sizes, orientation:
                calls.append((docks, sizes, orientation)),
        )
        fake_window._positive_int = self.main_window_module.MainWindow._positive_int

        self.main_window_module.MainWindow._restore_saved_video_dock_width(
            fake_window)

        self.assertEqual(calls, [
            ([files, video], [550, 450], Qt.Horizontal),
        ])

    def test_geometry_is_restored_before_show(self):
        calls = []
        fake_window = types.SimpleNamespace(
            saved_geometry=QByteArray(b"geometry"),
            restoreGeometry=lambda value: calls.append(value),
        )

        self.main_window_module.MainWindow._restore_saved_geometry_before_show(
            fake_window)

        self.assertEqual(calls, [QByteArray(b"geometry")])

    def test_active_custom_view_setter_does_not_shadow_reader(self):
        fake_window = types.SimpleNamespace()
        fake_window._active_custom_view_id = types.MethodType(
            self.main_window_module.MainWindow._active_custom_view_id,
            fake_window)
        fake_window._set_active_custom_view_id = types.MethodType(
            self.main_window_module.MainWindow._set_active_custom_view_id,
            fake_window)

        fake_window._set_active_custom_view_id("view-1")

        self.assertTrue(callable(fake_window._active_custom_view_id))
        self.assertEqual(fake_window._active_custom_view_id(), "view-1")
        self.assertEqual(self.app.settings.values["active_custom_view"], "view-1")
        self.assertEqual(self.app.settings.values["active_builtin_view"], "")

    def test_builtin_and_custom_view_identity_are_mutually_exclusive(self):
        fake_window = types.SimpleNamespace(_active_custom_view_id_value="view-1")

        self.main_window_module.MainWindow._set_active_builtin_view(fake_window, "recording")

        self.assertEqual(self.app.settings.values["active_builtin_view"], "recording")
        self.assertEqual(self.app.settings.values["active_custom_view"], "")
        self.assertEqual(fake_window._active_custom_view_id_value, "")

        self.main_window_module.MainWindow._set_active_custom_view_id(fake_window, "custom-1")

        self.assertEqual(self.app.settings.values["active_builtin_view"], "")
        self.assertEqual(self.app.settings.values["active_custom_view"], "custom-1")

    def test_scheduled_dock_style_update_waits_for_mouse_release(self):
        starts = []
        styles = []
        fake_window = types.SimpleNamespace(
            _dock_style_theme_changed=False,
            _dock_style_timer=types.SimpleNamespace(start=starts.append),
            style_dock_widgets=lambda theme_changed=False: styles.append(theme_changed),
        )

        with patch.object(
            self.main_window_module.QApplication,
            "mouseButtons",
            return_value=Qt.LeftButton,
        ):
            self.main_window_module.MainWindow._apply_scheduled_dock_style_update(fake_window)

        self.assertEqual(starts, [50])
        self.assertEqual(styles, [])

        fake_window._dock_style_theme_changed = True
        with patch.object(
            self.main_window_module.QApplication,
            "mouseButtons",
            return_value=Qt.NoButton,
        ):
            self.main_window_module.MainWindow._apply_scheduled_dock_style_update(fake_window)

        self.assertEqual(styles, [True])
        self.assertFalse(fake_window._dock_style_theme_changed)

    def test_save_project_emits_saved_signal_on_success(self):
        saved = SignalRecorder()
        failed = SignalRecorder()
        history_calls = []
        save_calls = []

        self.app.project = types.SimpleNamespace(save=save_calls.append)
        self.app.updates = types.SimpleNamespace(
            save_history=lambda project, limit: history_calls.append((project, limit))
        )
        self.app.settings.values["history-limit"] = 42

        fake_window = types.SimpleNamespace(
            lock=threading.Lock(),
            save_recovery=lambda path: save_calls.append(f"recovery:{path}"),
            ProjectSaved=saved,
            ProjectSaveFailed=failed,
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "project.osp")
            self.main_window_module.MainWindow.save_project(fake_window, project_path)

            self.assertEqual(history_calls, [(self.app.project, 42)])
            self.assertEqual(save_calls, [f"recovery:{project_path}", project_path])
            self.assertEqual(saved.calls, [(project_path,)])
            self.assertEqual(failed.calls, [])

    def test_optimized_preview_actions_use_cached_menu_targets_when_selection_is_empty(self):
        proxy_calls = []
        file_obj = types.SimpleNamespace(id="F1", data={"id": "F1", "media_type": "video"})
        fake_window = types.SimpleNamespace(
            selected_files=lambda: [],
            _optimized_preview_target_file_ids=["F1"],
            proxy_service=types.SimpleNamespace(
                remove_for_files=lambda files: proxy_calls.append(("remove", [getattr(f, "id", None) for f in files])),
                create_for_files=lambda files: proxy_calls.append(("create", [getattr(f, "id", None) for f in files])),
                use_existing_for_files=lambda files: proxy_calls.append(("locate", [getattr(f, "id", None) for f in files])),
                cancel_for_files=lambda files: proxy_calls.append(("cancel", [getattr(f, "id", None) for f in files])),
                delete_and_unlink_for_files=lambda files: proxy_calls.append(("delete", [getattr(f, "id", None) for f in files])),
            ),
        )
        fake_window._optimized_preview_files_for_action = lambda: self.main_window_module.MainWindow._optimized_preview_files_for_action(fake_window)

        with patch.object(self.main_window_module.File, "get", return_value=file_obj):
            files = self.main_window_module.MainWindow._optimized_preview_files_for_action(fake_window)
            self.assertEqual([f.id for f in files], ["F1"])

            self.main_window_module.MainWindow.actionOptimizedPreviewRemove_trigger(fake_window)
            self.main_window_module.MainWindow.actionOptimizedPreviewCreate_trigger(fake_window)
            self.main_window_module.MainWindow.actionOptimizedPreviewDeleteAndUnlink_trigger(fake_window)

        self.assertEqual(proxy_calls, [("remove", ["F1"]), ("create", ["F1"]), ("delete", ["F1"])])

    def test_optimized_preview_cancel_targets_current_file_only(self):
        proxy_calls = []
        file_obj = types.SimpleNamespace(id="F1", data={"id": "F1", "media_type": "video"})
        fake_window = types.SimpleNamespace(
            current_file_id=lambda: "F1",
            selected_files=lambda: [
                types.SimpleNamespace(id="F1", data={"id": "F1", "media_type": "video"}),
                types.SimpleNamespace(id="F2", data={"id": "F2", "media_type": "video"}),
                types.SimpleNamespace(id="F3", data={"id": "F3", "media_type": "video"}),
            ],
            _optimized_preview_target_file_ids=["F1", "F2", "F3"],
            proxy_service=types.SimpleNamespace(
                cancel_for_files=lambda files: proxy_calls.append(("cancel", [getattr(f, "id", None) for f in files])),
            ),
        )
        fake_window._optimized_preview_files_for_action = lambda: self.main_window_module.MainWindow._optimized_preview_files_for_action(fake_window)
        fake_window._optimized_preview_file_for_cancel_action = lambda: self.main_window_module.MainWindow._optimized_preview_file_for_cancel_action(fake_window)

        with patch.object(self.main_window_module.File, "get", return_value=file_obj):
            self.main_window_module.MainWindow.actionOptimizedPreviewCancel_trigger(fake_window)

        self.assertEqual(proxy_calls, [("cancel", ["F1"])])

    def test_open_project_missing_file_removes_recent_project_and_seeks_start(self):
        status_messages = []
        removed = []
        loaded_recent = []
        move_calls = []
        restore_cursor = []
        speed_calls = SignalRecorder()
        pause_calls = SignalRecorder()

        player = types.SimpleNamespace(Seek=lambda frame: move_calls.append(("seek", frame)))
        preview_thread = types.SimpleNamespace(player=player)
        video_preview = types.SimpleNamespace(
            clearTransformState=lambda: move_calls.append(("clear_transform",)),
            size=lambda: "preview-size",
        )

        fake_window = types.SimpleNamespace(
            SpeedSignal=speed_calls,
            PauseSignal=pause_calls,
            videoPreview=video_preview,
            clearSelections=lambda: move_calls.append(("clear_selections",)),
            statusBar=types.SimpleNamespace(showMessage=lambda text, ms: status_messages.append((text, ms))),
            remove_recent_project=removed.append,
            load_recent_menu=lambda: loaded_recent.append(True),
            movePlayhead=lambda frame: move_calls.append(("playhead", frame)),
            preview_thread=preview_thread,
            SetWindowTitle=lambda: None,
            refreshFilesSignal=SignalRecorder(),
            refreshFrameSignal=SignalRecorder(),
            MaxSizeChanged=SignalRecorder(),
            actionSave_trigger=lambda: None,
        )

        self.app.project = types.SimpleNamespace(needs_save=lambda: False)
        self.app.updates = types.SimpleNamespace(load_history=lambda project: None)
        self.app.window = fake_window
        self.app.setOverrideCursor = lambda cursor: None
        self.app.restoreOverrideCursor = lambda: restore_cursor.append(True)
        settings = openshot.Settings.Instance()
        previous_caching = settings.ENABLE_PLAYBACK_CACHING
        settings.ENABLE_PLAYBACK_CACHING = True

        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                missing_path = os.path.join(tmpdir, "missing.osp")
                with patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None):
                    self.main_window_module.MainWindow.open_project(
                        fake_window,
                        missing_path,
                        clear_thumbnails=True,
                    )

                self.assertEqual(removed, [missing_path])
                self.assertEqual(loaded_recent, [True])
                self.assertTrue(status_messages)
                self.assertIn("missing", status_messages[0][0].lower())
                self.assertIn(("seek", 1), move_calls)
                self.assertIn(("playhead", 1), move_calls)
                self.assertEqual(speed_calls.calls, [(0,)])
                self.assertEqual(pause_calls.calls, [()])
                self.assertEqual(restore_cursor, [True])
                self.assertTrue(settings.ENABLE_PLAYBACK_CACHING)
        finally:
            settings.ENABLE_PLAYBACK_CACHING = previous_caching

    def test_save_recovery_creates_zip_and_calls_retention(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "demo.osp")
            with open(project_path, "w", encoding="utf-8") as handle:
                handle.write("project-data")

            recovery_dir = os.path.join(tmpdir, "recovery")
            os.mkdir(recovery_dir)
            self.app.settings.values["recovery-limit"] = 10

            managed = []
            fake_window = types.SimpleNamespace(
                manage_recovery_files=lambda daily, historical, name: managed.append((daily, historical, name))
            )

            with ExitStack() as stack:
                stack.enter_context(
                    patch.object(self.main_window_module.info, "RECOVERY_PATH", recovery_dir)
                )
                stack.enter_context(
                    patch.object(self.main_window_module, "time", lambda: 1234567890)
                )
                self.main_window_module.MainWindow.save_recovery(fake_window, project_path)

            zip_path = os.path.join(recovery_dir, "1234567890-demo.zip")
            self.assertTrue(os.path.exists(zip_path))
            with zipfile.ZipFile(zip_path, "r") as archive:
                self.assertEqual(archive.namelist(), ["demo.osp"])
                self.assertEqual(archive.read("demo.osp").decode("utf-8"), "project-data")
            self.assertEqual(managed, [(7, 3, "demo")])

    def test_close_event_cancel_keeps_window_open(self):
        tutorial_calls = []
        save_calls = []
        event_calls = []

        fake_window = types.SimpleNamespace(
            tutorial_manager=types.SimpleNamespace(
                re_show_dialog=lambda: tutorial_calls.append("reshow"),
            ),
            actionSave_trigger=lambda: save_calls.append("save"),
            shutting_down=False,
        )
        self.app.project = types.SimpleNamespace(needs_save=lambda: True)

        event = types.SimpleNamespace(
            accept=lambda: event_calls.append("accept"),
            ignore=lambda: event_calls.append("ignore"),
        )

        with patch.object(
            self.main_window_module.QMessageBox,
            "question",
            return_value=self.main_window_module.QMessageBox.Cancel,
        ):
            self.main_window_module.MainWindow.closeEvent(fake_window, event)

        self.assertEqual(save_calls, [])
        self.assertEqual(tutorial_calls, ["reshow"])
        self.assertEqual(event_calls, ["ignore"])
        self.assertFalse(fake_window.shutting_down)

    def test_close_event_yes_saves_and_continues_shutdown(self):
        calls = []
        tracker = []
        event_calls = []

        fake_window = types.SimpleNamespace(
            tutorial_manager=None,
            actionSave_trigger=lambda: calls.append("save"),
            shutting_down=False,
            save_settings=lambda: calls.append("settings"),
            StopSignal=SignalRecorder(),
            http_server_thread=None,
            generation_queue=None,
            generation_service=None,
            preview_thread=None,
            preview_parent=None,
            videoPreview=None,
            timeline_sync=None,
            destroy_lock_file=lambda: calls.append("destroy_lock"),
        )
        self.app.project = types.SimpleNamespace(needs_save=lambda: True)
        self.app.logger_libopenshot = None
        event = types.SimpleNamespace(
            accept=lambda: event_calls.append("accept"),
            ignore=lambda: event_calls.append("ignore"),
        )

        with ExitStack() as stack:
            stack.enter_context(
                patch.object(
                    self.main_window_module.QMessageBox,
                    "question",
                    return_value=self.main_window_module.QMessageBox.Yes,
                )
            )
            stack.enter_context(
                patch.object(
                    self.main_window_module,
                    "track_metric_session",
                    tracker.append,
                )
            )
            stack.enter_context(
                patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None)
            )
            self.main_window_module.MainWindow.closeEvent(fake_window, event)

        self.assertEqual(event_calls, ["accept"])
        self.assertIn("save", calls)
        self.assertIn("settings", calls)
        self.assertIn("destroy_lock", calls)
        self.assertEqual(tracker, [False])
        self.assertTrue(fake_window.shutting_down)

    def test_clear_optimized_files_cancel_does_nothing(self):
        proxy_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                delete_internal_project_proxy_files=lambda: proxy_calls.append("delete"),
            ),
        )

        with patch.object(
            self.main_window_module.QMessageBox,
            "question",
            return_value=self.main_window_module.QMessageBox.No,
        ):
            self.main_window_module.MainWindow.actionClearOptimizedFiles_trigger(fake_window)

        self.assertEqual(proxy_calls, [])

    def test_clear_optimized_files_yes_deletes_project_optimized_files(self):
        proxy_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                delete_internal_project_proxy_files=lambda: proxy_calls.append("delete"),
            ),
        )

        with patch.object(
            self.main_window_module.QMessageBox,
            "question",
            return_value=self.main_window_module.QMessageBox.Yes,
        ):
            self.main_window_module.MainWindow.actionClearOptimizedFiles_trigger(fake_window)

        self.assertEqual(proxy_calls, ["delete"])

    def test_refresh_clear_menu_action_states_enables_action_only_when_internal_optimized_files_exist(self):
        enabled_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                has_internal_project_proxy_files=lambda: True,
            ),
            actionClearOptimizedFiles=types.SimpleNamespace(
                setEnabled=enabled_calls.append,
            ),
        )

        self.main_window_module.MainWindow._refresh_clear_menu_action_states(fake_window)

        self.assertEqual(enabled_calls, [True])

        enabled_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                has_internal_project_proxy_files=lambda: False,
            ),
            actionClearOptimizedFiles=types.SimpleNamespace(
                setEnabled=enabled_calls.append,
            ),
        )

        self.main_window_module.MainWindow._refresh_clear_menu_action_states(fake_window)

        self.assertEqual(enabled_calls, [False])

    def test_close_event_no_skips_save_but_shuts_down(self):
        calls = []

        fake_window = types.SimpleNamespace(
            tutorial_manager=None,
            actionSave_trigger=lambda: calls.append("save"),
            shutting_down=False,
            save_settings=lambda: calls.append("settings"),
            StopSignal=SignalRecorder(),
            http_server_thread=None,
            generation_queue=None,
            generation_service=None,
            preview_thread=None,
            preview_parent=None,
            videoPreview=None,
            timeline_sync=None,
            destroy_lock_file=lambda: calls.append("destroy_lock"),
        )
        self.app.project = types.SimpleNamespace(needs_save=lambda: True)
        self.app.logger_libopenshot = None
        event = types.SimpleNamespace(accept=lambda: None, ignore=lambda: None)

        with ExitStack() as stack:
            stack.enter_context(
                patch.object(
                    self.main_window_module.QMessageBox,
                    "question",
                    return_value=self.main_window_module.QMessageBox.No,
                )
            )
            stack.enter_context(
                patch.object(self.main_window_module, "track_metric_session", lambda value: None)
            )
            stack.enter_context(
                patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None)
            )
            self.main_window_module.MainWindow.closeEvent(fake_window, event)

        self.assertNotIn("save", calls)
        self.assertIn("settings", calls)
        self.assertIn("destroy_lock", calls)
        self.assertTrue(fake_window.shutting_down)

    def test_open_project_success_loads_project_and_refreshes_ui(self):
        refresh_files = SignalRecorder()
        refresh_frame = SignalRecorder()
        max_size = SignalRecorder()
        recent_calls = []
        clear_temp = []
        move_calls = []

        player = types.SimpleNamespace(Seek=lambda frame: move_calls.append(("seek", frame)))
        preview_thread = types.SimpleNamespace(player=player)
        video_preview = types.SimpleNamespace(
            clearTransformState=lambda: move_calls.append(("clear_transform",)),
            size=lambda: "preview-size",
        )

        fake_window = types.SimpleNamespace(
            SpeedSignal=SignalRecorder(),
            PauseSignal=SignalRecorder(),
            videoPreview=video_preview,
            clearSelections=lambda: move_calls.append(("clear_selections",)),
            preview_thread=preview_thread,
            SetWindowTitle=lambda: recent_calls.append("title"),
            refreshFilesSignal=refresh_files,
            refreshFrameSignal=refresh_frame,
            MaxSizeChanged=max_size,
            load_recent_menu=lambda: recent_calls.append("recent"),
            clear_temporary_files=lambda: clear_temp.append(True),
            movePlayhead=lambda frame: move_calls.append(("playhead", frame)),
            actionSave_trigger=lambda: None,
            statusBar=types.SimpleNamespace(showMessage=lambda *args: None),
            remove_recent_project=lambda path: None,
        )
        load_calls = []
        history_calls = []
        self.app.project = types.SimpleNamespace(
            needs_save=lambda: False,
            load=lambda path, clear_thumbnails: load_calls.append((path, clear_thumbnails)),
        )
        self.app.updates = types.SimpleNamespace(load_history=history_calls.append)
        self.app.window = fake_window
        self.app.setOverrideCursor = lambda cursor: None
        self.app.restoreOverrideCursor = lambda: recent_calls.append("restore")
        scheduled = []
        settings = openshot.Settings.Instance()
        previous_caching = settings.ENABLE_PLAYBACK_CACHING
        settings.ENABLE_PLAYBACK_CACHING = True

        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                project_path = os.path.join(tmpdir, "existing.osp")
                with ExitStack() as stack:
                    stack.enter_context(
                        patch.object(self.main_window_module.os.path, "exists", return_value=True)
                    )
                    stack.enter_context(
                        patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None)
                    )
                    stack.enter_context(
                        patch.object(
                            self.main_window_module.QTimer,
                            "singleShot",
                            side_effect=lambda delay, callback: scheduled.append((delay, callback)),
                        )
                    )
                    self.main_window_module.MainWindow.open_project(
                        fake_window,
                        project_path,
                        clear_thumbnails=True,
                    )

                self.assertEqual(load_calls, [(project_path, True)])
                self.assertEqual(history_calls, [self.app.project])
                self.assertEqual(clear_temp, [True])
                self.assertEqual(refresh_files.calls, [()])
                self.assertEqual(refresh_frame.calls, [])
                self.assertEqual(max_size.calls, [("preview-size",)])
                self.assertIn(("seek", 1), move_calls)
                self.assertFalse(settings.ENABLE_PLAYBACK_CACHING)
                self.assertEqual([delay for delay, _callback in scheduled], [0])
                scheduled[0][1]()
                self.assertEqual(refresh_frame.calls, [()])
        finally:
            settings.ENABLE_PLAYBACK_CACHING = previous_caching
        self.assertIn(("playhead", 1), move_calls)
        self.assertIn("recent", recent_calls)
        self.assertIn("restore", recent_calls)

    def test_action_remove_clip_skips_locked_tracks(self):
        deleted = []
        removed = []
        refreshed = SignalRecorder()

        locked_clip = types.SimpleNamespace(data={"layer": 2}, delete=lambda: deleted.append("locked"))
        unlocked_clip = types.SimpleNamespace(data={"layer": 1}, delete=lambda: deleted.append("unlocked"))

        self.app.project = types.SimpleNamespace(get=lambda key: [{"number": 2, "lock": True}])
        self.app.updates = types.SimpleNamespace(transaction_id=None)
        self.app.window = types.SimpleNamespace(refreshFrameSignal=refreshed)

        fake_window = types.SimpleNamespace(
            selected_clips=["C1"],
            removeSelection=lambda item_id, item_type: removed.append((item_id, item_type)),
            emit_selection_signal=lambda: None,
            show_property_timeout=lambda: None,
        )

        with patch.object(self.main_window_module.Clip, "filter", return_value=[locked_clip, unlocked_clip]):
            self.main_window_module.MainWindow.actionRemoveClip_trigger(fake_window, refresh=True)

        self.assertEqual(deleted, ["unlocked"])
        self.assertEqual(removed, [("C1", "clip")])
        self.assertEqual(refreshed.calls, [()])
        self.assertIsNone(self.app.updates.transaction_id)

    def test_action_remove_transition_skips_locked_tracks(self):
        deleted = []
        removed = []
        refreshed = SignalRecorder()

        locked_tran = types.SimpleNamespace(data={"layer": 3}, delete=lambda: deleted.append("locked"))
        unlocked_tran = types.SimpleNamespace(data={"layer": 1}, delete=lambda: deleted.append("unlocked"))

        self.app.project = types.SimpleNamespace(get=lambda key: [{"number": 3, "lock": True}])
        self.app.updates = types.SimpleNamespace(transaction_id=None)
        self.app.window = types.SimpleNamespace(refreshFrameSignal=refreshed)

        fake_window = types.SimpleNamespace(
            selected_transitions=["T1"],
            removeSelection=lambda item_id, item_type: removed.append((item_id, item_type)),
            emit_selection_signal=lambda: None,
            show_property_timeout=lambda: None,
            refreshFrameSignal=refreshed,
        )

        with patch.object(self.main_window_module.Transition, "filter", return_value=[locked_tran, unlocked_tran]):
            self.main_window_module.MainWindow.actionRemoveTransition_trigger(fake_window, refresh=True)

        self.assertEqual(deleted, ["unlocked"])
        self.assertEqual(removed, [("T1", "transition")])
        self.assertEqual(refreshed.calls, [()])
        self.assertIsNone(self.app.updates.transaction_id)

    def test_delete_item_removes_selected_effects(self):
        calls = []
        refreshed = SignalRecorder()

        self.app.updates = types.SimpleNamespace(transaction_id=None)

        fake_window = types.SimpleNamespace(
            filesView=types.SimpleNamespace(hasFocus=lambda: False),
            timeline=None,
            refreshFrameSignal=refreshed,
            actionRemoveEffect_trigger=lambda: calls.append("effect"),
            actionRemoveClip_trigger=lambda refresh=False: calls.append(("clip", refresh)),
            actionRemoveTransition_trigger=lambda refresh=False: calls.append(("transition", refresh)),
        )

        self.main_window_module.MainWindow.deleteItem(fake_window)

        self.assertEqual(calls, ["effect", ("clip", False), ("transition", False)])
        self.assertEqual(refreshed.calls, [()])
        self.assertIsNone(self.app.updates.transaction_id)

    def test_add_and_show_docks_keep_default_dock_features(self):
        fake_window = QMainWindow()
        normal_dock = QDockWidget("Normal", fake_window)
        normal_dock.setObjectName("dockNormal")

        self.main_window_module.MainWindow.addDocks(fake_window, [normal_dock], Qt.RightDockWidgetArea)
        self.assertTrue(normal_dock.features() & QDockWidget.DockWidgetClosable)
        self.assertTrue(normal_dock.features() & QDockWidget.DockWidgetMovable)
        self.assertTrue(normal_dock.features() & QDockWidget.DockWidgetFloatable)

        normal_dock.hide()
        fake_window.showDocks = lambda docks: self.main_window_module.MainWindow.showDocks(fake_window, docks)
        self.main_window_module.MainWindow.showDocks(fake_window, [normal_dock])
        self.assertTrue(normal_dock.features() & QDockWidget.DockWidgetClosable)
        self.assertTrue(normal_dock.features() & QDockWidget.DockWidgetMovable)
        self.assertTrue(normal_dock.features() & QDockWidget.DockWidgetFloatable)

    def test_properties_dock_reanchors_to_files_when_shown_after_view_removed_it(self):
        fake_window = QMainWindow()
        fake_window.dockFiles = QDockWidget("Project Files", fake_window)
        fake_window.dockFiles.setObjectName("dockFiles")
        fake_window.dockProperties = QDockWidget("Properties", fake_window)
        fake_window.dockProperties.setObjectName("dockProperties")
        fake_window.style_dock_widgets = lambda: None

        fake_window.addDockWidget(Qt.LeftDockWidgetArea, fake_window.dockFiles)
        fake_window.removeDockWidget(fake_window.dockProperties)

        self.main_window_module.MainWindow._anchor_and_show_properties_dock(fake_window)

        self.assertEqual(fake_window.dockWidgetArea(fake_window.dockProperties), Qt.LeftDockWidgetArea)
        self.assertIn(fake_window.dockProperties, fake_window.tabifiedDockWidgets(fake_window.dockFiles))
        self.assertFalse(fake_window.dockProperties.isFloating())
        self.assertFalse(fake_window.dockProperties.isHidden())

    def test_properties_dock_preserves_existing_simple_view_position(self):
        fake_window = QMainWindow()
        fake_window.dockFiles = QDockWidget("Project Files", fake_window)
        fake_window.dockFiles.setObjectName("dockFiles")
        fake_window.dockProperties = QDockWidget("Properties", fake_window)
        fake_window.dockProperties.setObjectName("dockProperties")
        fake_window.style_dock_widgets = lambda: None

        fake_window.addDockWidget(Qt.TopDockWidgetArea, fake_window.dockFiles)
        fake_window.addDockWidget(Qt.LeftDockWidgetArea, fake_window.dockProperties)
        fake_window.dockProperties.hide()

        self.main_window_module.MainWindow._anchor_and_show_properties_dock(fake_window)

        self.assertEqual(fake_window.dockWidgetArea(fake_window.dockProperties), Qt.LeftDockWidgetArea)
        self.assertNotIn(fake_window.dockProperties, fake_window.tabifiedDockWidgets(fake_window.dockFiles))
        self.assertFalse(fake_window.dockProperties.isHidden())

    def test_recording_view_matches_simple_layout_with_recording_on_right(self):
        fake_window = QMainWindow()
        fake_window.dockFiles = QDockWidget("Project Files", fake_window)
        fake_window.dockFiles.setObjectName("dockFiles")
        fake_window.dockTransitions = QDockWidget("Transitions", fake_window)
        fake_window.dockTransitions.setObjectName("dockTransitions")
        fake_window.dockEffects = QDockWidget("Effects", fake_window)
        fake_window.dockEffects.setObjectName("dockEffects")
        fake_window.dockEmojis = QDockWidget("Emojis", fake_window)
        fake_window.dockEmojis.setObjectName("dockEmojis")
        fake_window.dockProperties = QDockWidget("Properties", fake_window)
        fake_window.dockProperties.setObjectName("dockProperties")
        fake_window.dockVideo = QDockWidget("Video Preview", fake_window)
        fake_window.dockVideo.setObjectName("dockVideo")
        fake_window.dockTimeline = QDockWidget("Timeline", fake_window)
        fake_window.dockTimeline.setObjectName("dockTimeline")
        fake_window.dockAudioRecording = QDockWidget("Recording", fake_window)
        fake_window.dockAudioRecording.setObjectName("dockAudioRecording")
        fake_window._set_active_custom_view_id = lambda _view_id: None
        fake_window._set_active_builtin_view = lambda _view_id: None
        fake_window._ensure_audio_recording_dock_content = lambda: None
        fake_window._apply_simple_view_layout = lambda: (
            self.main_window_module.MainWindow._apply_simple_view_layout(fake_window)
        )
        fake_window.getDocks = lambda: fake_window.findChildren(QDockWidget)
        fake_window.removeDocks = lambda: self.main_window_module.MainWindow.removeDocks(fake_window)
        fake_window.addDocks = lambda docks, area: self.main_window_module.MainWindow.addDocks(fake_window, docks, area)
        fake_window.floatDocks = lambda is_floating: self.main_window_module.MainWindow.floatDocks(fake_window, is_floating)
        fake_window.showDocks = lambda docks: self.main_window_module.MainWindow.showDocks(fake_window, docks)
        fake_window.style_dock_widgets = lambda: None

        self.main_window_module.MainWindow.actionAudio_Recording_View_trigger(fake_window)

        self.assertEqual(fake_window.dockWidgetArea(fake_window.dockAudioRecording), Qt.RightDockWidgetArea)
        self.assertIn(fake_window.dockTransitions, fake_window.tabifiedDockWidgets(fake_window.dockFiles))
        self.assertIn(fake_window.dockEffects, fake_window.tabifiedDockWidgets(fake_window.dockFiles))
        self.assertIn(fake_window.dockEmojis, fake_window.tabifiedDockWidgets(fake_window.dockFiles))
        self.assertFalse(fake_window.dockFiles.isHidden())
        self.assertFalse(fake_window.dockVideo.isHidden())
        self.assertFalse(fake_window.dockTimeline.isHidden())

    def test_scope_menu_keeps_conditional_show_and_close_all_actions(self):
        fake_window = QMainWindow()
        fake_window.scopes_menu = QMenu(fake_window)
        fake_window.dockAudio = QDockWidget("Audio Levels", fake_window)
        fake_window.dockAudio.setObjectName("dockAudio")
        fake_window.dockHistogram = QDockWidget("Histogram", fake_window)
        fake_window.dockHistogram.setObjectName("dockHistogram")
        fake_window.dockLumaWaveform = QDockWidget("Luma Waveform", fake_window)
        fake_window.dockLumaWaveform.setObjectName("dockLumaWaveform")
        fake_window.dockVectorscope = QDockWidget("Vectorscope", fake_window)
        fake_window.dockVectorscope.setObjectName("dockVectorscope")
        for dock in [
                fake_window.dockAudio,
                fake_window.dockHistogram,
                fake_window.dockLumaWaveform,
                fake_window.dockVectorscope]:
            fake_window.addDockWidget(Qt.RightDockWidgetArea, dock)
            dock.hide()

        open_docks = set()
        fake_window._scope_docks = lambda: self.main_window_module.MainWindow._scope_docks(fake_window)
        fake_window._dock_is_open = lambda dock: dock in open_docks
        fake_window.closeDocks = lambda docks: self.main_window_module.MainWindow.closeDocks(fake_window, docks)
        fake_window.show_all_scope_docks = lambda: None
        fake_window._add_dock_visibility_actions = (
            lambda menu, docks, show_text, close_text, show_callback=None:
            self.main_window_module.MainWindow._add_dock_visibility_actions(
                fake_window, menu, docks, show_text, close_text, show_callback))

        self.main_window_module.MainWindow._rebuild_scopes_menu(fake_window)
        action_texts = [action.text() for action in fake_window.scopes_menu.actions() if not action.isSeparator()]
        self.assertIn("Show All Scopes", action_texts)
        self.assertNotIn("Close All Scopes", action_texts)
        self.assertNotIn("Lock Scopes", action_texts)

        open_docks.add(fake_window.dockAudio)
        self.main_window_module.MainWindow._rebuild_scopes_menu(fake_window)
        action_texts = [action.text() for action in fake_window.scopes_menu.actions() if not action.isSeparator()]
        self.assertIn("Show All Scopes", action_texts)
        self.assertIn("Close All Scopes", action_texts)
        self.assertNotIn("Unlock Scopes", action_texts)

    def test_live_property_resume_keeps_cache_disabled_until_seek_or_play(self):
        settings = openshot.Settings.Instance()
        previous = settings.ENABLE_PLAYBACK_CACHING
        try:
            settings.ENABLE_PLAYBACK_CACHING = False
            fake_view = types.SimpleNamespace(live_property_cache_paused=True)

            self.properties_tableview_module.PropertiesTableView.resume_live_property_caching(fake_view)

            self.assertFalse(fake_view.live_property_cache_paused)
            self.assertFalse(settings.ENABLE_PLAYBACK_CACHING)
        finally:
            settings.ENABLE_PLAYBACK_CACHING = previous

    def test_insert_keyframe_adds_current_color_property_frame(self):
        saved = []
        refreshed = SignalRecorder()
        self.app.window = types.SimpleNamespace(refreshFrameSignal=refreshed)

        effect = types.SimpleNamespace(
            data={
                "wave_color": {
                    "red": {"Points": [{"co": {"X": 1.0, "Y": 0.0}, "interpolation": openshot.LINEAR}]},
                    "green": {"Points": [{"co": {"X": 1.0, "Y": 123.0}, "interpolation": openshot.LINEAR}]},
                    "blue": {"Points": [{"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR}]},
                    "alpha": {"Points": [{"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR}]},
                }
            },
        )
        effect.save = lambda: saved.append(effect.data)

        model = QStandardItemModel()
        label = QStandardItem("Wave Color")
        label.setData((
            "wave_color",
            {
                "type": "color",
                "red": {"value": 0},
                "green": {"value": 123},
                "blue": {"value": 255},
                "alpha": {"value": 255},
                "closest_point_x": 1,
                "previous_point_x": 1,
                "object_id": None,
                "max": 255.0,
            },
        ))
        value = QStandardItem("")
        value.setData([("effect-1", "effect")])
        model.appendRow([label, value])

        parent = types.SimpleNamespace(
            currentIndex=lambda: model.index(0, 0),
            clearSelection=lambda: None,
            setCurrentIndex=lambda index: None,
        )
        helper = self.properties_model_module.PropertiesModel.__new__(
            self.properties_model_module.PropertiesModel)
        helper.model = model
        helper.parent = parent
        helper.frame_number = 30
        helper._trim_preview_mode = False

        with patch.object(self.properties_model_module.Effect, "get", return_value=effect):
            helper.insert_keyframe(value)

        self.assertEqual(len(saved), 1)
        color = effect.data["wave_color"]
        self.assertIn(30, [point["co"]["X"] for point in color["red"]["Points"]])
        self.assertIn(30, [point["co"]["X"] for point in color["green"]["Points"]])
        self.assertIn(30, [point["co"]["X"] for point in color["blue"]["Points"]])
        self.assertIn(30, [point["co"]["X"] for point in color["alpha"]["Points"]])
        self.assertEqual(refreshed.calls, [()])

    def test_ripple_delete_gap_shifts_only_later_items_on_same_layer(self):
        saved = []
        clips = [
            types.SimpleNamespace(data={"position": 4.0}, save=lambda: saved.append("clip-before")),
            types.SimpleNamespace(data={"position": 9.0}, save=lambda: saved.append("clip-after")),
        ]
        transitions = [
            types.SimpleNamespace(data={"position": 8.0}, save=lambda: saved.append("tran-before")),
            types.SimpleNamespace(data={"position": 12.0}, save=lambda: saved.append("tran-after")),
        ]

        with ExitStack() as stack:
            stack.enter_context(patch.object(self.main_window_module.Clip, "filter", return_value=clips))
            stack.enter_context(
                patch.object(self.main_window_module.Transition, "filter", return_value=transitions)
            )
            self.main_window_module.MainWindow.ripple_delete_gap(
                types.SimpleNamespace(),
                ripple_start=8.0,
                layer=1,
                total_gap=2.5,
            )

        self.assertEqual(clips[0].data["position"], 4.0)
        self.assertEqual(clips[1].data["position"], 6.5)
        self.assertEqual(transitions[0].data["position"], 8.0)
        self.assertEqual(transitions[1].data["position"], 9.5)
        self.assertEqual(saved, ["clip-after", "tran-after"])



#!/usr/bin/python3
"""
 @file
 @brief This file verifies all translations are correctly formatted and have the correct # of string replacements
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import argparse
import ast
import shutil
import os
import re
import fnmatch
import subprocess
import sys
from collections import Counter
from collections import Counter
from typing import Any, Dict, List, Optional, Tuple

SRC_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
if SRC_PATH not in sys.path:
    sys.path.insert(0, SRC_PATH)

from qt_api import QCoreApplication, QTranslator


# Absolute path of the translations directory
LANG_PATH = os.path.dirname(os.path.abspath(__file__))
REPO_PATH = os.path.dirname(os.path.dirname(LANG_PATH))
DOC_LOCALE_PATH = os.path.join(REPO_PATH, 'doc', 'locale')

# Match '%(name)x' format placeholders
TAG_RE = re.compile(r'%\(([^\)]*)\)(.)')
DOC_SUB_RE = re.compile(r'\|[A-Za-z0-9_]+\|')
PRINTF_TOKEN_RE = re.compile(
    r'%(?:'
    r'%'  # escaped percent literal
    r'|\(([^)]+)\)[#0\-+]*\d*(?:\.\d+)?[diouxXeEfFgGcrsa]'  # named placeholder
    r'|[#0\-+]*\d*(?:\.\d+)?[diouxXeEfFgGcrsa]'  # positional/unnamed placeholder
    r')'
)

# Match numbers in strings (for detecting number-embedded strings like "Line 1")
NUMBER_RE = re.compile(r'\d+')


class BadTranslationsError(Exception):
    pass


class POEntry:
    """Minimal PO entry representation for doc validation."""

    def __init__(self) -> None:
        self.msgid = ""
        self.msgid_plural = ""
        self.msgstrs: Dict[int, str] = {}
        self.references: List[str] = []


def get_doc_locale_from_path(path: str) -> Optional[str]:
    """Extract the doc locale code from a PO file path."""
    parts = os.path.normpath(path).split(os.sep)
    try:
        locale_index = parts.index('locale')
    except ValueError:
        return None
    if locale_index + 1 >= len(parts):
        return None
    return parts[locale_index + 1]


class Color:
    """Color for message output"""
    _red = '\u001b[31m'
    _yellow = '\u001b[33m'
    _green = '\u001b[32m'
    _reset = '\u001b[0m'

    @classmethod
    def red(cls, *args: Optional[Any]) -> str:
        return cls._red + str(*args) + cls._reset

    @classmethod
    def yellow(cls, *args: Optional[Any]) -> str:
        return cls._yellow + str(*args) + cls._reset

    @classmethod
    def green(cls, *args: Optional[Any]) -> str:
        return cls._green + str(*args) + cls._reset


def get_app() -> QCoreApplication:
    """Create the translation app lazily so test discovery doesn't lock in QCoreApplication."""
    app = QCoreApplication.instance()
    if app is not None:
        return app
    return QCoreApplication(sys.argv)


def build_stringlists() -> Dict[str, List]:
    """ Create a dict containing lists of strings, keyed on source filename"""
    all_strings = {}

    for pot_file in [
        'OpenShot.pot',
        'OpenShot_transitions.pot',
        'OpenShot_blender.pot',
        'OpenShot_emojis.pot',
    ]:
        with open(os.path.join(LANG_PATH, 'OpenShot', pot_file)) as f:
            data = f.read()
        all_strings.update({
            pot_file: re.findall('^msgid \"(.*)\"', data, re.MULTILINE)
        })
    return all_strings


def check_trans(strings: List) -> List[Tuple[str, str]]:
    """Check all strings in a list against a given .qm file"""
    app = get_app()
    # Test translation of all strings
    translations = {
        source: app.translate("", source)
        for source in strings
    }
    # Check for replacements with mismatched number of % escapes
    errors = {}
    for s, t in translations.items():
        # Count placeholders in source and translation
        s_count = s.count('%s') + s.count('%d') + s.count('%f')
        t_count = t.count('%s') + t.count('%d') + t.count('%f')

        # If counts match, no error
        if s.count('%s') == t.count('%s') and \
           s.count('%d') == t.count('%d') and \
           s.count('%f') == t.count('%f'):
            continue

        # Special case: source has embedded numbers but no placeholders,
        # and translation has %s placeholders instead.
        # This is valid when translators use a placeholder pattern
        # (e.g., "Line 1" translated as "Linia %s" where %s = "1")
        source_numbers = NUMBER_RE.findall(s)
        if source_numbers and s_count == 0 and t_count == len(source_numbers):
            continue

        errors[s] = t
    # Check for missing/added variable names
    # e.g.: "%(clip_id)s %(value)d" changed to "%(clip)s %(value)d"
    # or mismatched types
    # e.g.: "%(seconds)s" changed to "%(seconds)d"
    named_variables = {
        s: (TAG_RE.findall(s), TAG_RE.findall(t))
        for s, t in translations.items()
        if s.count('%(') > 0
    }
    errors.update({
        s: translations[s]
        for s, (s_vars, t_vars) in named_variables.items()
        # Ensure that t_vars is a strict subset of s_vars
        if not set(t_vars) <= set(s_vars)
    })
    return list(errors.items())


def process_qm(file: str, stringlists: Dict[str, List[str]]) -> int:
    """Scan a translation file against all provided strings"""
    app = get_app()
    # Attempt to load translation file
    basename = os.path.splitext(file)[0]
    translator = QTranslator(app)
    if not translator.load(basename, LANG_PATH):
        print(Color.red('QTranslator failed to load') + f' {file}')
        return 1

    app.installTranslator(translator)

    # Build a dict mapping source POTfiles to lists of error pairs
    error_sets = {
        sourcefile: check_trans(strings)
        for sourcefile, strings in stringlists.items()
    }

    app.removeTranslator(translator)

    # Display any errors found, grouped by source POT file
    error_count = sum([len(v) for v in error_sets.values()])
    invalid_msg = "Invalid"
    if error_count:
        print(f'{file}: ' + Color.red(f'{error_count} total errors'))
    for pot, errset in error_sets.items():
        if not errset:
            continue
        width = len(pot)
        for source, trans in errset:
            print(Color.yellow(f'{pot}:') + f' {source}')
            print(Color.red(f'{invalid_msg:>{width}}:') + f' {trans}\n')
    return error_count


def scan_all_qm(filenames: List[str] = None) -> None:
    all_strings = build_stringlists()
    if not filenames:
        filenames = fnmatch.filter(os.listdir(LANG_PATH), 'OpenShot*.qm')
    # Loop through language files and count errors
    total_errors = sum([
        process_qm(filename, all_strings)
        for filename in filenames
    ])

    string_count = sum([len(s) for s in all_strings.values()])
    lang_count = len(filenames)

    print(f"Tested {Color.yellow(string_count)} strings on "
          + f"{Color.yellow(lang_count)} translation files.")
    if total_errors > 0:
        raise BadTranslationsError(f"Found {total_errors} translation errors")


def po_unquote(text: str) -> str:
    """Decode a quoted PO string token."""
    return ast.literal_eval(text)


def extract_percent_tokens(text: str) -> Tuple[Counter, Counter]:
    """Extract printf-style percent tokens."""
    tokens = Counter()
    named_tokens = Counter()

    for match in PRINTF_TOKEN_RE.finditer(text):
        token = match.group(0)
        tokens[token] += 1
        named_match = TAG_RE.match(token)
        if named_match:
            named_tokens[named_match.groups()] += 1

    return tokens, named_tokens


def extract_doc_substitutions(text: str) -> Counter:
    """Extract Sphinx substitution tokens like |icon_name|."""
    return Counter(DOC_SUB_RE.findall(text))


def flush_po_entry(entries: List[POEntry], entry: POEntry) -> POEntry:
    """Append the current PO entry if it contains any data, then start a new one."""
    if entry.msgid or entry.msgstrs or entry.msgid_plural or entry.references:
        entries.append(entry)
    return POEntry()


def parse_po_file(path: str) -> List[POEntry]:
    """Parse enough of a PO file to inspect msgid/msgstr tokens."""
    entries: List[POEntry] = []
    entry = POEntry()
    active_field: Optional[Tuple[str, Optional[int]]] = None

    with open(path, encoding='utf-8') as handle:
        for raw_line in handle:
            line = raw_line.rstrip('\n')

            if not line.strip():
                entry = flush_po_entry(entries, entry)
                active_field = None
                continue

            if line.startswith('#~'):
                continue

            if line.startswith('#:'):
                entry.references.extend(line[2:].strip().split())
                continue

            if line.startswith('#'):
                continue

            if line.startswith('msgid_plural '):
                entry.msgid_plural = po_unquote(line[len('msgid_plural '):])
                active_field = ('msgid_plural', None)
                continue

            if line.startswith('msgid '):
                if entry.msgid or entry.msgstrs or entry.msgid_plural:
                    entry = flush_po_entry(entries, entry)
                entry.msgid = po_unquote(line[len('msgid '):])
                active_field = ('msgid', None)
                continue

            if line.startswith('msgstr['):
                index_text, value = line.split(']', 1)
                index = int(index_text[len('msgstr['):])
                entry.msgstrs[index] = po_unquote(value.strip()[1:])
                active_field = ('msgstr', index)
                continue

            if line.startswith('msgstr '):
                entry.msgstrs[0] = po_unquote(line[len('msgstr '):])
                active_field = ('msgstr', 0)
                continue

            if line.startswith('msgctxt '):
                active_field = ('msgctxt', None)
                continue

            if line.startswith('"'):
                value = po_unquote(line)
                if active_field == ('msgid', None):
                    entry.msgid += value
                elif active_field == ('msgid_plural', None):
                    entry.msgid_plural += value
                elif active_field and active_field[0] == 'msgstr' and active_field[1] is not None:
                    entry.msgstrs[active_field[1]] = entry.msgstrs.get(active_field[1], '') + value
                continue

    flush_po_entry(entries, entry)
    return entries


def validate_doc_entry(path: str, entry: POEntry) -> List[str]:
    """Check doc translation entries for placeholder and substitution preservation."""
    if entry.msgid == "":
        header = entry.msgstrs.get(0, '')
        header_fields: Dict[str, str] = {}
        for line in header.splitlines():
            if ':' not in line:
                continue
            key, value = line.split(':', 1)
            header_fields[key.strip()] = value.strip()

        expected_language = get_doc_locale_from_path(path)
        actual_language = header_fields.get('Language', '')
        if not expected_language:
            return []

        errors = []
        if not actual_language:
            errors.append(f"{path}: PO header is missing a Language value")
        elif actual_language != expected_language:
            errors.append(
                f"{path}: PO header Language '{actual_language}' does not match locale '{expected_language}'"
            )
        return errors

    source_strings = [entry.msgid]
    if entry.msgid_plural:
        source_strings.append(entry.msgid_plural)

    expected_named = Counter()
    expected_tokens = Counter()
    expected_subs = Counter()
    for source in source_strings:
        source_tokens, source_named = extract_percent_tokens(source)
        expected_named |= source_named
        expected_tokens |= source_tokens
        expected_subs |= extract_doc_substitutions(source)

    errors = []
    for plural_index, translation in sorted(entry.msgstrs.items()):
        if not translation.strip():
            continue

        actual_tokens, actual_named = extract_percent_tokens(translation)
        actual_subs = extract_doc_substitutions(translation)

        label = path
        if entry.references:
            label += f" ({', '.join(entry.references[:2])})"
        if len(entry.msgstrs) > 1:
            label += f" [msgstr[{plural_index}]]"

        if actual_named != expected_named:
            errors.append(
                f"{label}: named placeholders differ\n"
                f"  msgid: {entry.msgid}\n"
                f"  msgstr: {translation}"
            )

        if actual_tokens != expected_tokens:
            errors.append(
                f"{label}: printf tokens differ\n"
                f"  msgid: {entry.msgid}\n"
                f"  msgstr: {translation}"
            )

        if actual_subs != expected_subs:
            errors.append(
                f"{label}: Sphinx substitution tokens differ\n"
                f"  msgid: {entry.msgid}\n"
                f"  msgstr: {translation}"
            )

    return errors


def process_doc_po(path: str) -> int:
    """Run syntax and token-preservation checks on a doc PO file."""
    errors: List[str] = []
    msgfmt_path = shutil.which('msgfmt')
    trusted_path = os.path.abspath(path)
    trusted_roots = (os.path.abspath(LANG_PATH), os.path.abspath(DOC_LOCALE_PATH))

    if not msgfmt_path:
        errors.append(f"{path}: msgfmt executable not found")
        return len(errors)

    if not any(
        os.path.commonpath([trusted_path, root]) == root
        for root in trusted_roots
    ):
        errors.append(f"{path}: path is outside trusted locale roots")
        return len(errors)

    msgfmt_args = [
        msgfmt_path,
        '--check',
        '--check-format',
        '--output-file=/dev/null',
        trusted_path,
    ]
    msgfmt = subprocess.run(
        msgfmt_args,
        capture_output=True,
        text=True,
        check=False,
    )
    if msgfmt.returncode != 0:
        stderr = msgfmt.stderr.strip() or msgfmt.stdout.strip() or 'msgfmt validation failed'
        errors.append(f"{path}: {stderr}")

    try:
        entries = parse_po_file(path)
    except Exception as ex:
        errors.append(f"{path}: PO parse failed: {ex}")
        entries = []

    for entry in entries:
        errors.extend(validate_doc_entry(path, entry))

    if errors:
        print(Color.red(f'{path}: {len(errors)} total errors'))
        for error in errors:
            print(error + '\n')
    return len(errors)


def scan_doc_po_files() -> None:
    """Validate all documentation PO files."""
    po_files = []
    for root, _, files in os.walk(DOC_LOCALE_PATH):
        for filename in files:
            if filename.endswith('.po'):
                po_files.append(os.path.join(root, filename))
    po_files.sort()

    total_errors = sum(process_doc_po(path) for path in po_files)
    print(f"Tested {Color.yellow(len(po_files))} documentation PO files.")
    if total_errors > 0:
        raise BadTranslationsError(f"Found {total_errors} documentation translation errors")


def parse_args(argv: List[str]) -> argparse.Namespace:
    """Parse CLI arguments without breaking the existing filename-only behavior."""
    parser = argparse.ArgumentParser(description='Validate OpenShot translation files.')
    parser.add_argument('filenames', nargs='*', help='Optional .qm files to validate')
    parser.add_argument('--docs', action='store_true', help='Validate doc gettext PO files')
    parser.add_argument('--all', action='store_true', help='Validate both app and doc translations')
    return parser.parse_args(argv)


# Autorun if used as script
if __name__ == '__main__':
    try:
        args = parse_args(sys.argv[1:])
        if args.all:
            scan_all_qm(args.filenames)
            scan_doc_po_files()
        elif args.docs:
            scan_doc_po_files()
        else:
            scan_all_qm(args.filenames)
    except BadTranslationsError as ex:
        print(Color.red(str(ex) + "! See above."))
        exit(1)
    else:
        print(Color.green("No errors found."))
        exit(0)



"""
 @file
 @brief Shared helpers for best-effort HTTP requests in packaged OpenShot builds
 @author OpenShot Studios, LLC

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 """

import os
import ssl
import urllib.request
from urllib.parse import urlparse, urlunparse

import requests

from classes import info
from classes.logger import log


DEFAULT_CONNECT_TIMEOUT = 2
DEFAULT_READ_TIMEOUT = 5
DOWNLOAD_READ_SIZE = 1024 * 1024
URLLIB_ALLOWED_SCHEMES = ("http", "https")


def ca_bundle_path():
    """Return the packaged CA bundle path when available."""
    try:
        import certifi

        cafile = certifi.where()
        if cafile and os.path.exists(cafile):
            return cafile
    except Exception as ex:
        log.debug("Unable to locate certifi CA bundle: %s", ex)
    return None


def configure_ssl_environment():
    """Point common Python HTTP clients at certifi when packaged."""
    cafile = ca_bundle_path()
    if not cafile:
        log.info("No packaged CA bundle detected; using Python/OpenSSL default CA paths")
        return None

    os.environ.setdefault("SSL_CERT_FILE", cafile)
    os.environ.setdefault("REQUESTS_CA_BUNDLE", cafile)
    log.info("Configured packaged CA bundle: %s", cafile)
    return cafile


def ssl_context():
    """Return an SSL context using the packaged CA bundle when available."""
    cafile = ca_bundle_path()
    if cafile:
        return ssl.create_default_context(cafile=cafile)
    return ssl.create_default_context()


def verify_path():
    """Return a requests-compatible verify value."""
    return ca_bundle_path() or True


def http_fallback_url(url):
    """Return an HTTP fallback URL for an HTTPS URL."""
    parsed = urlparse(url)
    if parsed.scheme.lower() != "https":
        return None
    return urlunparse(parsed._replace(scheme="http"))


def validate_url_scheme(url, allowed_schemes=URLLIB_ALLOWED_SCHEMES):
    """Validate URLs before passing them to urllib helpers."""
    scheme = urlparse(str(url)).scheme.lower()
    if scheme not in allowed_schemes:
        raise ValueError("Unsupported URL scheme: {}".format(scheme or "<empty>"))
    return scheme


def urls_with_http_fallback(url):
    """Return URL attempts in preferred order."""
    urls = [url]
    fallback = http_fallback_url(url)
    if fallback and fallback not in urls:
        urls.append(fallback)
    return urls


def get_json(urls, description, headers=None, timeout=None):
    """Fetch JSON from one or more URLs, logging each fallback attempt."""
    if isinstance(urls, str):
        urls = [urls]
    timeout = timeout or (DEFAULT_CONNECT_TIMEOUT, DEFAULT_READ_TIMEOUT)
    headers = headers or {}
    last_error = None

    for index, url in enumerate(urls, start=1):
        try:
            log.info("HTTP GET %s attempt %s/%s: %s", description, index, len(urls), url)
            response = requests.get(
                url,
                headers=headers,
                timeout=timeout,
                verify=verify_path(),
            )
            response.raise_for_status()
            log.info("HTTP GET %s succeeded via %s", description, urlparse(url).scheme.upper())
            return response.json()
        except Exception as ex:
            last_error = ex
            log.warning("HTTP GET %s failed via %s: %s", description, url, ex)

    raise RuntimeError("Failed to fetch %s from %s URL(s): %s" % (description, len(urls), last_error))


def post_json(url, payload, description, headers=None, timeout=None):
    """POST JSON with packaged CA handling and logging."""
    timeout = timeout or (DEFAULT_CONNECT_TIMEOUT, DEFAULT_READ_TIMEOUT)
    headers = headers or {}
    log.info("HTTP POST %s: %s", description, url)
    response = requests.post(
        url,
        json=payload,
        headers=headers,
        timeout=timeout,
        verify=verify_path(),
    )
    response.raise_for_status()
    log.info("HTTP POST %s succeeded with status=%s", description, response.status_code)
    return response


def download_file(urls, output_path, description, progress_callback=None, timeout=None, cancel_exceptions=None):
    """Download one of several URL attempts to disk."""
    if isinstance(urls, str):
        urls = [urls]
    timeout = timeout or (DEFAULT_CONNECT_TIMEOUT, DEFAULT_READ_TIMEOUT)
    cancel_exceptions = tuple(cancel_exceptions or ())
    last_error = None

    for index, url in enumerate(urls, start=1):
        try:
            log.info("HTTP download %s attempt %s/%s: %s", description, index, len(urls), url)
            _download_file_once(url, output_path, progress_callback, timeout)
            log.info("HTTP download %s succeeded via %s", description, urlparse(url).scheme.upper())
            return url
        except Exception as ex:
            if cancel_exceptions and isinstance(ex, cancel_exceptions):
                log.info("HTTP download %s cancelled", description)
                raise
            last_error = ex
            log.warning("HTTP download %s failed via %s: %s", description, url, ex)
            try:
                if os.path.exists(output_path):
                    os.remove(output_path)
            except OSError:
                log.warning("Failed to remove partial download: %s", output_path, exc_info=True)

    raise RuntimeError("Failed to download %s from %s URL(s): %s" % (description, len(urls), last_error))


def _download_file_once(url, output_path, progress_callback, timeout):
    scheme = validate_url_scheme(url)
    request = urllib.request.Request(
        url,
        headers={"User-Agent": "OpenShot/{}".format(info.VERSION)},
    )
    kwargs = {"timeout": sum(timeout) if isinstance(timeout, tuple) else timeout}
    if scheme == "https":
        kwargs["context"] = ssl_context()

    with urllib.request.urlopen(request, **kwargs) as response:  # nosec B310 - URL scheme is restricted above.
        total_size = response.headers.get("Content-Length")
        total_size = int(total_size) if total_size else 0
        downloaded_size = 0

        with open(output_path, "wb") as output_file:
            while True:
                chunk = response.read(DOWNLOAD_READ_SIZE)
                if not chunk:
                    break

                output_file.write(chunk)
                downloaded_size += len(chunk)
                if progress_callback:
                    progress_callback(downloaded_size, total_size)



"""
 @file
 @brief Unit tests for camera motion framing helpers
 @author OpenShot Studios

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 """

import os
import sys
import unittest


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes.camera_motion import (
    KEN_BURNS_AUTO,
    KEN_BURNS_LEFT_TO_RIGHT,
    KEN_BURNS_TOP_TO_BOTTOM,
    PAN_AUTO,
    PAN_BOTTOM_TO_TOP,
    PAN_RIGHT,
    PAN_LEFT_TO_RIGHT,
    PAN_UP,
    camera_pan_keyframes,
    ken_burns_keyframes,
    push_pull_keyframes,
    source_dimensions_from_reader,
)


class CameraMotionTests(unittest.TestCase):
    def test_wide_media_pans_horizontally_without_extra_zoom(self):
        values = camera_pan_keyframes(PAN_LEFT_TO_RIGHT, 1920, 1080, 3840, 1080)

        self.assertEqual(values.scale_x, (1.0, 1.0))
        self.assertEqual(values.scale_y, (1.0, 1.0))
        self.assertGreater(values.location_x[0], 0.0)
        self.assertLess(values.location_x[1], 0.0)
        self.assertEqual(values.location_y, (0.0, 0.0))

    def test_tall_media_pans_vertically_without_extra_zoom(self):
        values = camera_pan_keyframes(PAN_UP, 1920, 1080, 1080, 3840)

        self.assertEqual(values.scale_x, (1.0, 1.0))
        self.assertEqual(values.scale_y, (1.0, 1.0))
        self.assertLess(values.location_y[0], 0.0)
        self.assertGreater(values.location_y[1], 0.0)
        self.assertGreater(abs(values.location_y[0]), 0.4)
        self.assertEqual(values.location_x, (0.0, 0.0))

    def test_tall_two_by_three_image_pans_to_crop_edges(self):
        values = camera_pan_keyframes(PAN_BOTTOM_TO_TOP, 1920, 1080, 1024, 1536)

        self.assertEqual(values.scale_y, (1.0, 1.0))
        self.assertAlmostEqual(abs(values.location_y[0]), 0.452272, places=5)
        self.assertAlmostEqual(abs(values.location_y[1]), 0.452272, places=5)

    def test_auto_pan_chooses_natural_direction(self):
        values = camera_pan_keyframes(PAN_AUTO, 1920, 1080, 1024, 1536)

        self.assertEqual(values.location_x, (0.0, 0.0))
        self.assertLess(values.location_y[0], 0.0)
        self.assertGreater(values.location_y[1], 0.0)

    def test_cross_axis_pan_adds_only_needed_zoom(self):
        values = camera_pan_keyframes(PAN_RIGHT, 1920, 1080, 1080, 1920)

        self.assertGreater(values.scale_x[0], 1.0)
        self.assertEqual(values.scale_x, values.scale_y)
        self.assertGreater(values.location_x[0], 0.0)
        self.assertLess(values.location_x[1], 0.0)
        self.assertEqual(values.location_y, (0.0, 0.0))

    def test_push_pull_are_centered_zoom_only(self):
        push = push_pull_keyframes(zoom_in=True)
        pull = push_pull_keyframes(zoom_in=False)

        self.assertEqual(push.scale_x[0], 1.0)
        self.assertGreater(push.scale_x[1], 1.0)
        self.assertGreater(pull.scale_x[0], 1.0)
        self.assertEqual(pull.scale_x[1], 1.0)
        self.assertEqual(push.location_x, (0.0, 0.0))
        self.assertEqual(pull.location_y, (0.0, 0.0))

    def test_auto_ken_burns_chooses_wide_axis(self):
        values = ken_burns_keyframes(True, KEN_BURNS_AUTO, 1920, 1080, 3840, 1080)

        self.assertEqual(values.scale_x[0], 1.0)
        self.assertGreater(values.scale_x[1], 1.0)
        self.assertGreater(values.location_x[0], 0.0)
        self.assertLess(values.location_x[1], 0.0)
        self.assertEqual(values.location_y, (0.0, 0.0))

    def test_auto_ken_burns_chooses_tall_axis(self):
        values = ken_burns_keyframes(True, KEN_BURNS_AUTO, 1920, 1080, 1080, 3840)

        self.assertEqual(values.location_x, (0.0, 0.0))
        self.assertLess(values.location_y[0], 0.0)
        self.assertGreater(values.location_y[1], 0.0)

    def test_forced_ken_burns_direction_uses_requested_axis(self):
        values = ken_burns_keyframes(True, KEN_BURNS_TOP_TO_BOTTOM, 1920, 1080, 1080, 3840)

        self.assertEqual(values.location_x, (0.0, 0.0))
        self.assertGreater(values.location_y[0], 0.0)
        self.assertLess(values.location_y[1], 0.0)

    def test_ken_burns_out_reverses_zoom_but_keeps_requested_travel(self):
        values = ken_burns_keyframes(False, KEN_BURNS_LEFT_TO_RIGHT, 1920, 1080, 3840, 1080)

        self.assertGreater(values.scale_x[0], 1.0)
        self.assertEqual(values.scale_x[1], 1.0)
        self.assertGreater(values.location_x[0], 0.0)
        self.assertLess(values.location_x[1], 0.0)

    def test_source_dimensions_from_reader_accepts_display_dimensions(self):
        self.assertEqual(
            source_dimensions_from_reader({"display_width": 800, "display_height": 600}),
            (800.0, 600.0),
        )
        self.assertEqual(source_dimensions_from_reader({}), (None, None))


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Unit tests for shared HTTP helpers
 @author OpenShot Studios, LLC

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.
 """

import os
import sys
import types
import unittest
from unittest.mock import patch


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes import http_client


class HttpClientTests(unittest.TestCase):
    def test_urls_with_http_fallback_prefers_https(self):
        self.assertEqual(
            http_client.urls_with_http_fallback("https://example.com/file.zip"),
            ["https://example.com/file.zip", "http://example.com/file.zip"],
        )

    def test_get_json_falls_back_to_http_after_https_failure(self):
        class Response:
            status_code = 200

            def raise_for_status(self):
                return None

            def json(self):
                return {"ok": True}

        with patch.object(
            http_client.requests,
            "get",
            side_effect=[RuntimeError("ssl failed"), Response()],
        ) as get:
            result = http_client.get_json(
                ["https://example.com/version/json/", "http://example.com/version/json/"],
                "test metadata",
            )

        self.assertEqual(result, {"ok": True})
        self.assertEqual(get.call_args_list[0].args[0], "https://example.com/version/json/")
        self.assertEqual(get.call_args_list[1].args[0], "http://example.com/version/json/")

    def test_download_file_does_not_fallback_after_cancel(self):
        class Cancelled(Exception):
            pass

        def progress_callback(_downloaded, _total):
            raise Cancelled()

        class Response:
            headers = {"Content-Length": "4"}

            def __enter__(self):
                return self

            def __exit__(self, *_args):
                return False

            def read(self, _size):
                return b"data"

        with patch.object(http_client, "ssl_context", return_value=object()), \
                patch.object(http_client.urllib.request, "urlopen", return_value=Response()) as urlopen:
            with self.assertRaises(Cancelled):
                http_client.download_file(
                    ["https://example.com/file.zip", "http://example.com/file.zip"],
                    os.devnull,
                    "test download",
                    progress_callback,
                    cancel_exceptions=(Cancelled,),
                )

        urlopen.assert_called_once()

    def test_configure_ssl_environment_sets_certifi_paths(self):
        certifi_stub = types.SimpleNamespace(where=lambda: "/tmp/cacert.pem")
        with patch.dict(sys.modules, {"certifi": certifi_stub}), \
                patch("classes.http_client.os.path.exists", return_value=True), \
                patch.dict(os.environ, {}, clear=True):
            self.assertEqual(http_client.configure_ssl_environment(), "/tmp/cacert.pem")
            self.assertEqual(os.environ["SSL_CERT_FILE"], "/tmp/cacert.pem")
            self.assertEqual(os.environ["REQUESTS_CA_BUNDLE"], "/tmp/cacert.pem")


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Scope dock panels: waveform, histogram, vectorscope, and audio meters.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import math
import os

from qt_api import (
    Qt, pyqtSignal, pyqtSlot,
    QWidget, QVBoxLayout, QHBoxLayout, QSizePolicy,
    QImage, QPainter, QColor, QPen, QBrush, QIcon,
    QComboBox, QToolButton, QRect, QPainterPath, QPointF,
)
from classes import info
from classes.logger import log
from windows.color_grade_editor import draw_broadcast_hue_ring

# ─── Persistent settings keys ────────────────────────────────────────────────
_S_WAVE_MODE  = "scope-waveform-mode"     # luma|red|green|blue|rgb_overlay|rgb_parade
_S_WAVE_COLOR = "scope-waveform-color"    # green|white|orange
_S_HIST_CH    = "scope-histogram-channel" # rgba|luma|red|green|blue
_S_HIST_SCALE = "scope-histogram-scale"   # log|linear
_S_VEC_DISPLAY = "scope-vectorscope-display"  # colorized|density|intensity
_S_VEC_ZOOM    = "scope-vectorscope-zoom"     # 100|200|400

_VECTORSCOPE_HUE_LABELS = (
    ("R", 108.65, 360.0),
    ("Mg", 51.65, 300.0),
    ("B", 350.76, 240.0),
    ("Cy", 288.65, 180.0),
    ("G", 231.65, 120.0),
    ("Yi", 170.76, 60.0 + 360.0),
)
_vectorscope_geometry_cache = {}
_vectorscope_intensity_lut = None
_vectorscope_label_lut_cache = {}   # size → (6, size*size) float32 numpy array


def N_(message):
    """Mark a deferred UI string for translation without translating at import time."""
    return message


def _settings():
    try:
        from classes.app import get_app
        return get_app().get_settings()
    except Exception:
        return None


def _get(key, default):
    s = _settings()
    if s is None:
        return default
    v = s.get(key)
    return v if v is not None else default


def _set(key, value):
    s = _settings()
    if s is not None:
        try:
            s.set(key, value)
        except Exception as ex:
            log.debug("Unable to save scope setting %s: %s", key, ex)


def _scope_region_icon(size=16):
    icon_path = os.path.join(info.PATH, "themes", "cosmic", "images", "tool-scope-region.svg")
    return QIcon(icon_path)


def _make_scope_region_button(parent):
    from classes.app import get_app
    _ = get_app()._tr
    button = QToolButton(parent)
    button.setCheckable(True)
    button.setAutoRaise(True)
    button.setIcon(_scope_region_icon())
    button.setToolTip(_("Analyze selected preview region"))
    button.setFocusPolicy(Qt.NoFocus)
    return button


def normalize_vectorscope_display(value):
    if value == "monochrome":
        return "density"
    if value == "heatmap":
        return "intensity"
    return value or "colorized"


def _vectorscope_wrap_angle(angle):
    return (angle + 360.0) % 360.0


def _vectorscope_display_hue_for_scope_angle(angle_deg):
    anchors = sorted((_vectorscope_wrap_angle(scope_angle), display_hue)
                     for _, scope_angle, display_hue in _VECTORSCOPE_HUE_LABELS)
    wrapped = _vectorscope_wrap_angle(angle_deg)
    normalized = []
    previous_hue = None
    for scope_angle, display_hue in anchors:
        hue = display_hue
        if previous_hue is not None:
            while hue < previous_hue:
                hue += 360.0
        normalized.append((scope_angle, hue))
        previous_hue = hue
    if wrapped < normalized[0][0]:
        wrapped += 360.0
    extended = normalized + [(normalized[0][0] + 360.0, normalized[0][1] + 360.0)]

    for index in range(len(normalized)):
        a0, h0 = extended[index]
        a1, h1 = extended[index + 1]
        if a0 <= wrapped <= a1:
            span = max(1e-6, a1 - a0)
            t = (wrapped - a0) / span
            return (h0 + ((h1 - h0) * t)) % 360.0

    return normalized[0][1] % 360.0


def _vectorscope_density_to_byte(count, max_val):
    if count <= 0 or max_val <= 0:
        return 0
    return min(255, int(math.sqrt(count / max_val) * 255))


def _build_vectorscope_label_lut(size):
    """Build a (6, size*size) float32 weight matrix — computed once per size."""
    import numpy as np
    half_cone = 45.0
    center = (size - 1) * 0.5
    ix = np.arange(size, dtype=np.float32)
    iy = np.arange(size, dtype=np.float32)
    u_norm = (ix[np.newaxis, :] - center) / center   # (1,size) → broadcast (size,size)
    v_norm = (center - iy[:, np.newaxis]) / center   # (size,1) → broadcast
    radius = np.minimum(np.sqrt(u_norm ** 2 + v_norm ** 2), 1.0)
    angle_deg = np.degrees(np.arctan2(v_norm, u_norm))
    lut = np.empty((6, size * size), dtype=np.float32)
    for i, (_, scope_angle, _) in enumerate(_VECTORSCOPE_HUE_LABELS):
        diff = np.abs(((angle_deg - scope_angle + 180.0) % 360.0) - 180.0)
        weight = np.where(diff < half_cone,
                          np.cos(np.radians(diff * (90.0 / half_cone))), 0.0)
        lut[i] = (weight * radius).ravel().astype(np.float32)
    return lut


def _vectorscope_label_scores(flat_density, size):
    """Return 6 proximity scores [0..1] for each HUE_LABEL, one dot-product away."""
    try:
        import numpy as np
    except ImportError:
        return [0.0] * 6
    if not flat_density or len(flat_density) != size * size:
        return [0.0] * 6
    lut = _vectorscope_label_lut_cache.get(size)
    if lut is None:
        lut = _build_vectorscope_label_lut(size)
        _vectorscope_label_lut_cache[size] = lut
    arr = np.asarray(flat_density, dtype=np.float32)
    total = float(arr.sum())
    if total <= 0.0:
        return [0.0] * 6
    scores = lut @ arr        # (6,) weighted pixel sums
    scores /= total           # average weight per pixel → nominally 0..1
    np.sqrt(scores, out=scores)  # sqrt compresses dynamic range for better feel
    np.clip(scores, 0.0, 1.0, out=scores)
    return scores.tolist()


def _vectorscope_geometry_info(size, zoom_factor):
    key = (int(size), int(round(zoom_factor * 100)))
    cached = _vectorscope_geometry_cache.get(key)
    if cached is not None:
        return cached

    source_indices = [-1] * (size * size)
    colorized_base = bytearray(size * size * 3)
    center = (size - 1) * 0.5
    radius = center
    inv_zoom = 1.0 / max(1.0, zoom_factor)

    for y in range(size):
        row_offset = y * size
        source_y = int(round(center + ((y - center) * inv_zoom)))
        for x in range(size):
            source_x = int(round(center + ((x - center) * inv_zoom)))
            if source_x < 0 or source_x >= size or source_y < 0 or source_y >= size:
                continue

            pixel_index = row_offset + x
            source_indices[pixel_index] = (source_y * size) + source_x

            u = (x - center) / max(1.0, radius)
            v = (center - y) / max(1.0, radius)
            hue = _vectorscope_display_hue_for_scope_angle(math.degrees(math.atan2(v, u)))
            saturation = min(1.0, math.sqrt((u * u) + (v * v)))
            color = QColor.fromHsv(int(hue) % 360, int(255 * saturation), 255)
            color_index = pixel_index * 3
            colorized_base[color_index] = color.red()
            colorized_base[color_index + 1] = color.green()
            colorized_base[color_index + 2] = color.blue()

    cached = (source_indices, colorized_base)
    _vectorscope_geometry_cache[key] = cached
    return cached


def _vectorscope_intensity_lut_bytes():
    global _vectorscope_intensity_lut
    if _vectorscope_intensity_lut is not None:
        return _vectorscope_intensity_lut

    lut = bytearray(256 * 3)
    for t in range(256):
        hue = int(max(0, 240 - (t * 240 / 255)))
        saturation = min(255, 180 + (t * 75 // 255))
        value = min(255, 56 + (t * 199 // 255))
        color = QColor.fromHsv(hue, saturation, value)
        idx = t * 3
        lut[idx] = color.red()
        lut[idx + 1] = color.green()
        lut[idx + 2] = color.blue()
    _vectorscope_intensity_lut = lut
    return _vectorscope_intensity_lut


def build_vectorscope_image(flat, size, zoom_factor, display):
    display = normalize_vectorscope_display(display)
    if not flat or len(flat) != size * size:
        return None

    max_val = max(flat) or 1
    buf = bytearray(size * size * 3)
    source_indices, colorized_base = _vectorscope_geometry_info(size, zoom_factor)
    intensity_lut = _vectorscope_intensity_lut_bytes() if display == "intensity" else None

    for pixel_index, source_index in enumerate(source_indices):
        if source_index < 0:
            continue

        count = flat[source_index]
        if not count:
            continue

        t = _vectorscope_density_to_byte(count, max_val)
        idx = pixel_index * 3
        if display == "density":
            value = min(255, 24 + (t * 231 // 255))
            buf[idx] = value
            buf[idx + 1] = value
            buf[idx + 2] = value
        elif display == "intensity":
            lut_index = t * 3
            buf[idx] = intensity_lut[lut_index]
            buf[idx + 1] = intensity_lut[lut_index + 1]
            buf[idx + 2] = intensity_lut[lut_index + 2]
        else:
            value = min(255, 80 + (t * 175 // 255))
            buf[idx] = (colorized_base[idx] * value) // 255
            buf[idx + 1] = (colorized_base[idx + 1] * value) // 255
            buf[idx + 2] = (colorized_base[idx + 2] * value) // 255
    return QImage(bytes(buf), size, size, size * 3, QImage.Format_RGB888).copy()


# ─── Waveform painter ────────────────────────────────────────────────────────

class WaveformWidget(QWidget):
    """Luma / RGB waveform density heatmap painter."""

    _LUMA_COLORS = {
        "green":  (0,   220,  80),
        "white":  (220, 220, 220),
        "orange": (255, 160,   0),
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self._data  = None
        self._mode  = _get(_S_WAVE_MODE,  "luma")
        self._color = _get(_S_WAVE_COLOR, "green")
        self._ire   = True
        self.setMinimumSize(120, 100)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setFocusPolicy(Qt.NoFocus)

    def set_mode(self,  v): self._mode  = v; _set(_S_WAVE_MODE,  v); self.update()
    def set_color(self, v): self._color = v; _set(_S_WAVE_COLOR, v); self.update()

    def update_data(self, video_data):
        self._data = video_data
        self.update()

    # ── helpers ──────────────────────────────────────────────────────────────

    @staticmethod
    def _density_to_byte(count, max_val):
        """Boost low-density samples so sparse scopes stay visible."""
        if count <= 0 or max_val <= 0:
            return 0
        return min(255, int(math.sqrt(count / max_val) * 255))

    def _build_img(self, flat, columns, bins, rgb):
        """Build a columns×bins RGB888 QImage from a waveform flat array."""
        if not flat or len(flat) != columns * bins:
            return None
        max_val = max(flat) or 1
        r0, g0, b0 = rgb
        buf = bytearray(columns * bins * 3)
        for col in range(columns):
            base = col * bins
            for b in range(bins):
                count = flat[base + b]
                if not count:
                    continue
                t = self._density_to_byte(count, max_val)
                row = bins - 1 - b
                idx = (row * columns + col) * 3
                buf[idx]     = r0 * t // 255
                buf[idx + 1] = g0 * t // 255
                buf[idx + 2] = b0 * t // 255
        # Detach from the temporary Python buffer so paint reads stable pixels.
        return QImage(bytes(buf), columns, bins, columns * 3, QImage.Format_RGB888).copy()

    # ── paint ────────────────────────────────────────────────────────────────

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, False)
        painter.setRenderHint(QPainter.SmoothPixmapTransform, True)
        w, h = self.width(), self.height()
        painter.fillRect(0, 0, w, h, QColor(0, 0, 0))

        if not self._data or not self._data.get("present"):
            painter.setPen(QColor(80, 80, 80))
            painter.drawText(self.rect(), Qt.AlignCenter, "No Video")
            return

        wf      = self._data.get("waveform", {})
        columns = wf.get("columns", 256)
        bins    = wf.get("bins",    256)
        mode    = self._mode

        if mode == "rgb_parade":
            band_w = w // 3
            for i, (key, rgb) in enumerate([
                ("red",   (220,  60,  60)),
                ("green", ( 60, 190,  60)),
                ("blue",  ( 60, 120, 220)),
            ]):
                img = self._build_img(wf.get(key, []), columns, bins, rgb)
                if img:
                    painter.drawImage(QRect(i * band_w, 0, band_w, h), img)

            # Thin dividers between bands
            painter.setPen(QPen(QColor(50, 50, 50), 1))
            painter.drawLine(band_w,     0, band_w,     h)
            painter.drawLine(band_w * 2, 0, band_w * 2, h)

        elif mode == "rgb_overlay":
            r_flat = wf.get("red",   [])
            g_flat = wf.get("green", [])
            b_flat = wf.get("blue",  [])
            if r_flat and len(r_flat) == columns * bins:
                max_val = max(max(r_flat), max(g_flat) if g_flat else 0,
                              max(b_flat) if b_flat else 0) or 1
                buf = bytearray(columns * bins * 3)
                for col in range(columns):
                    base = col * bins
                    for b in range(bins):
                        rv = r_flat[base + b]
                        gv = g_flat[base + b] if g_flat else 0
                        bv = b_flat[base + b] if b_flat else 0
                        if not (rv or gv or bv):
                            continue
                        row = bins - 1 - b
                        idx = (row * columns + col) * 3
                        buf[idx]     = min(255, self._density_to_byte(rv, max_val) * 220 // 255)
                        buf[idx + 1] = min(255, self._density_to_byte(gv, max_val) * 190 // 255)
                        buf[idx + 2] = min(255, self._density_to_byte(bv, max_val) * 220 // 255)
                img = QImage(bytes(buf), columns, bins, columns * 3,
                             QImage.Format_RGB888).copy()
                painter.drawImage(self.rect(), img)

        else:
            # Single-channel modes: luma, red, green, blue
            rgb_map = {
                "luma":  self._LUMA_COLORS.get(self._color, (0, 220, 80)),
                "red":   (220,  60,  60),
                "green": ( 60, 190,  60),
                "blue":  ( 60, 120, 220),
            }
            flat = wf.get(mode if mode != "luma" else "luma", [])
            img  = self._build_img(flat, columns, bins, rgb_map.get(mode, (200, 200, 200)))
            if img:
                painter.drawImage(self.rect(), img)

        # IRE reference lines at 10 / 50 / 90 %
        if self._ire:
            painter.setPen(QPen(QColor(60, 60, 60), 1, Qt.DashLine))
            for pct in (0.1, 0.5, 0.9):
                y = int(h * (1.0 - pct))
                painter.drawLine(0, y, w, y)


# ─── Histogram painter ───────────────────────────────────────────────────────

class HistogramWidget(QWidget):
    """RGB + luma overlay histogram with channel and scale filters."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._data    = None
        self._channel = _get(_S_HIST_CH,    "rgba")
        self._scale   = _get(_S_HIST_SCALE, "log")
        self.setMinimumSize(120, 80)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setFocusPolicy(Qt.NoFocus)

    def set_channel(self, v): self._channel = v; _set(_S_HIST_CH,    v); self.update()
    def set_scale(self,   v): self._scale   = v; _set(_S_HIST_SCALE, v); self.update()

    def update_data(self, video_data):
        self._data = video_data
        self.update()

    @staticmethod
    def _bin_span(index, bins, width):
        """Map one source bin to an exact on-screen span without gaps."""
        x0 = index * width // bins
        x1 = (index + 1) * width // bins
        if x1 <= x0:
            x1 = min(width, x0 + 1)
        return x0, x1

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, False)
        w, h = self.width(), self.height()
        painter.fillRect(0, 0, w, h, QColor(20, 20, 20))

        if not self._data or not self._data.get("present"):
            painter.setPen(QColor(80, 80, 80))
            painter.drawText(self.rect(), Qt.AlignCenter, "No Video")
            return

        hist  = self._data.get("histogram", {})
        luma  = hist.get("luma",  [])
        red   = hist.get("red",   [])
        green = hist.get("green", [])
        blue  = hist.get("blue",  [])

        if not luma:
            return

        ch = self._channel
        if ch == "rgba":
            to_draw = [
                (blue,  QColor(60,  60,  220, 140)),
                (green, QColor(60,  190,  60, 140)),
                (red,   QColor(220,  60,  60, 140)),
                (luma,  QColor(210, 210, 210,  70)),
            ]
        elif ch == "luma":
            to_draw = [(luma,  QColor(210, 210, 210, 200))]
        elif ch == "red":
            to_draw = [(red,   QColor(220,  60,  60, 200))]
        elif ch == "green":
            to_draw = [(green, QColor( 60, 190,  60, 200))]
        elif ch == "blue":
            to_draw = [(blue,  QColor( 60,  60, 220, 200))]
        else:
            return

        all_vals = [v for vals, _ in to_draw for v in vals]
        max_val  = max(all_vals) if all_vals else 1
        if not max_val:
            max_val = 1
        use_log  = (self._scale == "log")
        log_max  = math.log1p(max_val)
        bins     = len(luma)

        painter.setPen(Qt.NoPen)
        if ch == "rgba":
            painter.setCompositionMode(QPainter.CompositionMode_Plus)
        for vals, color in to_draw:
            painter.setBrush(QBrush(color))
            for i, v in enumerate(vals):
                if not v:
                    continue
                x, x2 = self._bin_span(i, bins, w)
                bar_h = int(math.log1p(v) / log_max * h) if use_log else v * h // max_val
                painter.drawRect(x, h - bar_h, x2 - x, bar_h)
        painter.setCompositionMode(QPainter.CompositionMode_SourceOver)


# ─── Vectorscope painter ─────────────────────────────────────────────────────

class VectorscopeWidget(QWidget):
    """2D chroma density plot with a lightweight vectorscope graticule."""

    _HUE_LABELS = _VECTORSCOPE_HUE_LABELS
    _SKIN_TONE_ANGLE = 123.0

    def __init__(self, parent=None):
        super().__init__(parent)
        self._data = None
        self._display = normalize_vectorscope_display(_get(_S_VEC_DISPLAY, "colorized"))
        self._zoom = _get(_S_VEC_ZOOM, "100")
        self._render_token = 0
        self._cached_scope_image = None
        self._cached_scope_key = None
        self._cached_overlay_image = None
        self._cached_overlay_key = None
        self._label_scores = [0.0] * 6
        self.setMinimumSize(120, 100)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setFocusPolicy(Qt.NoFocus)

    def set_display(self, value):
        self._display = normalize_vectorscope_display(value)
        self._render_token += 1
        self._cached_scope_image = None
        self._cached_scope_key = None
        _set(_S_VEC_DISPLAY, self._display)
        self.update()

    def set_zoom(self, value):
        self._zoom = value
        self._render_token += 1
        self._cached_scope_image = None
        self._cached_scope_key = None
        _set(_S_VEC_ZOOM, value)
        self.update()

    def update_data(self, video_data):
        self._data = video_data
        self._render_token += 1
        self._cached_scope_image = None
        self._cached_scope_key = None
        self.update()

    def resizeEvent(self, event):
        self._cached_scope_image = None
        self._cached_scope_key = None
        self._cached_overlay_image = None
        self._cached_overlay_key = None
        super().resizeEvent(event)

    def _build_img(self, flat, size, zoom_factor):
        return build_vectorscope_image(flat, size, zoom_factor, self._display)

    def _draw_guides(self, painter, center_x, center_y, radius):
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing, True)
        ring_pen = QPen(QColor(90, 98, 110, 88), 1)
        spoke_pen = QPen(QColor(90, 98, 110, 68), 1)
        skin_pen = QPen(QColor(210, 170, 120, 110), 1, Qt.DashLine)

        painter.setPen(ring_pen)
        for step in range(1, 7):
            r = radius * step / 6.0
            painter.drawEllipse(QRect(
                int(center_x - r), int(center_y - r),
                int(r * 2), int(r * 2),
            ))

        painter.setPen(spoke_pen)
        for _, angle_deg, _ in self._HUE_LABELS:
            angle = math.radians(angle_deg)
            x = center_x + (math.cos(angle) * radius)
            y = center_y - (math.sin(angle) * radius)
            painter.drawLine(int(center_x), int(center_y), int(x), int(y))

        painter.drawLine(int(center_x - radius), int(center_y), int(center_x + radius), int(center_y))
        painter.drawLine(int(center_x), int(center_y - radius), int(center_x), int(center_y + radius))

        painter.setPen(skin_pen)
        skin_angle = math.radians(self._SKIN_TONE_ANGLE)
        skin_x = center_x + (math.cos(skin_angle) * radius)
        skin_y = center_y - (math.sin(skin_angle) * radius)
        painter.drawLine(int(center_x), int(center_y), int(skin_x), int(skin_y))

        painter.restore()

    def _draw_labels(self, painter, center_x, center_y, radius, scores):
        """Draw hue labels per-frame, tinted toward their target color by proximity score."""
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing, True)
        base_r, base_g, base_b, base_a = 160, 168, 176, 120
        for i, (label, angle_deg, display_hue) in enumerate(self._HUE_LABELS):
            score = scores[i] if scores else 0.0
            angle = math.radians(angle_deg)
            lx = center_x + (math.cos(angle) * radius * 0.82)
            ly = center_y - (math.sin(angle) * radius * 0.82)
            if score < 0.02:
                color = QColor(base_r, base_g, base_b, base_a)
            else:
                target = QColor.fromHsvF((display_hue % 360.0) / 360.0, 1.0, 1.0)
                r = int(base_r + (target.red()   - base_r) * score)
                g = int(base_g + (target.green() - base_g) * score)
                b = int(base_b + (target.blue()  - base_b) * score)
                a = int(base_a + (255 - base_a) * score)
                color = QColor(r, g, b, a)
            painter.setPen(QPen(color, 1))
            painter.drawText(QRect(int(lx - 18), int(ly - 10), 36, 20), Qt.AlignCenter, label)
        painter.restore()

    def _layout_info(self):
        w, h = self.width(), self.height()
        side = max(10, min(w, h) - 12)
        left = (w - side) // 2
        top = (h - side) // 2
        center_x = left + side / 2.0
        center_y = top + side / 2.0
        radius = side / 2.0
        ring_margin = max(8.0, radius * 0.075)
        plot_radius = max(8.0, radius - ring_margin)
        plot_left = center_x - plot_radius
        plot_top = center_y - plot_radius
        plot_size = plot_radius * 2.0
        plot_rect = QRect(
            int(plot_left),
            int(plot_top),
            int(plot_size),
            int(plot_size),
        )
        return {
            "widget_width": w,
            "widget_height": h,
            "center_x": center_x,
            "center_y": center_y,
            "radius": radius,
            "plot_radius": plot_radius,
            "plot_left": plot_left,
            "plot_top": plot_top,
            "plot_size": plot_size,
            "plot_rect": plot_rect,
        }

    def _ensure_overlay_cache(self, layout):
        key = (
            layout["widget_width"],
            layout["widget_height"],
            int(layout["center_x"] * 10),
            int(layout["center_y"] * 10),
            int(layout["radius"] * 10),
            int(layout["plot_radius"] * 10),
        )
        if self._cached_overlay_key == key and self._cached_overlay_image is not None:
            return

        overlay = QImage(layout["widget_width"], layout["widget_height"], QImage.Format_ARGB32_Premultiplied)
        overlay.fill(QColor(0, 0, 0, 0))
        painter = QPainter(overlay)
        try:
            painter.setRenderHint(QPainter.Antialiasing, True)
            self._draw_guides(painter, layout["center_x"], layout["center_y"], layout["plot_radius"])
            draw_broadcast_hue_ring(
                painter,
                QPointF(layout["center_x"], layout["center_y"]),
                layout["radius"] - 2.0,
                3.0,
                alpha=220,
            )
            painter.setPen(QPen(QColor(120, 128, 138, 110), 1))
            painter.drawEllipse(layout["plot_rect"])
        finally:
            painter.end()

        self._cached_overlay_image = overlay
        self._cached_overlay_key = key

    def _ensure_scope_cache(self, layout):
        vectorscope = self._data.get("vectorscope", {}) if self._data else {}
        size = vectorscope.get("size", 256)
        flat = vectorscope.get("density", [])
        zoom_factor = {"100": 1.0, "200": 2.0, "400": 4.0}.get(str(self._zoom), 1.0)
        key = (
            self._render_token,
            layout["plot_rect"].width(),
            layout["plot_rect"].height(),
            self._display,
            self._zoom,
            size,
            len(flat),
        )
        if self._cached_scope_key == key and self._cached_scope_image is not None:
            return

        self._cached_scope_image = None
        self._cached_scope_key = key
        self._label_scores = _vectorscope_label_scores(flat, size)
        img = vectorscope.get("image")
        if not isinstance(img, QImage) or img.isNull():
            img = self._build_img(flat, size, zoom_factor)
        if img is None:
            return

        scope_image = QImage(layout["widget_width"], layout["widget_height"], QImage.Format_ARGB32_Premultiplied)
        scope_image.fill(QColor(0, 0, 0, 0))
        painter = QPainter(scope_image)
        try:
            painter.setRenderHint(QPainter.SmoothPixmapTransform, True)
            clip_path = QPainterPath()
            clip_path.addEllipse(layout["plot_left"], layout["plot_top"], layout["plot_size"], layout["plot_size"])
            painter.setClipPath(clip_path)
            painter.drawImage(layout["plot_rect"], img)
        finally:
            painter.end()
        self._cached_scope_image = scope_image

    def paintEvent(self, event):
        painter = QPainter(self)
        try:
            painter.setRenderHint(QPainter.Antialiasing, True)
            painter.setRenderHint(QPainter.SmoothPixmapTransform, True)
            w, h = self.width(), self.height()
            painter.fillRect(0, 0, w, h, QColor(0, 0, 0))

            if not self._data or not self._data.get("present"):
                painter.setPen(QColor(80, 80, 80))
                painter.drawText(self.rect(), Qt.AlignCenter, "No Video")
                return

            layout = self._layout_info()
            self._ensure_scope_cache(layout)
            self._ensure_overlay_cache(layout)

            if self._cached_scope_image is not None:
                painter.drawImage(0, 0, self._cached_scope_image)
            if self._cached_overlay_image is not None:
                painter.drawImage(0, 0, self._cached_overlay_image)
            self._draw_labels(painter, layout["center_x"], layout["center_y"],
                              layout["plot_radius"], self._label_scores)
        finally:
            painter.end()


# ─── Audio meter painter ─────────────────────────────────────────────────────

class AudioMeterWidget(QWidget):
    """Per-channel RMS/peak VU bars with clip indicator."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._data = None
        self.setMinimumSize(60, 80)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setFocusPolicy(Qt.NoFocus)

    def update_data(self, audio_data):
        self._data = audio_data
        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, False)
        w, h = self.width(), self.height()
        painter.fillRect(0, 0, w, h, QColor(20, 20, 20))

        if not self._data or not self._data.get("present"):
            painter.setPen(QColor(80, 80, 80))
            painter.drawText(self.rect(), Qt.AlignCenter, "No Audio")
            return

        summary   = self._data.get("summary", {})
        peak_vals = summary.get("peak", [])
        rms_vals  = summary.get("rms", [])
        clipped   = summary.get("clipped_samples", [])
        channels  = self._data.get("channels", 0)

        if not channels:
            return

        gap   = 4
        bar_w = max(8, (w - (channels + 1) * gap) // channels)

        for ch in range(channels):
            x             = gap + ch * (bar_w + gap)
            rms           = rms_vals[ch]  if ch < len(rms_vals)  else 0.0
            peak          = peak_vals[ch] if ch < len(peak_vals) else 0.0
            clipped_count = clipped[ch]   if ch < len(clipped)   else 0

            painter.fillRect(x, 0, bar_w, h, QColor(40, 40, 40))

            rms_h = int(rms * h)
            for row in range(rms_h):
                ratio = row / h
                if ratio > 0.8:
                    color = QColor(220, 50, 50)
                elif ratio > 0.6:
                    color = QColor(220, 200, 50)
                else:
                    color = QColor(50, 200, 50)
                painter.fillRect(x, h - row - 1, bar_w, 1, color)

            if peak > 0:
                peak_y = int((1.0 - peak) * h)
                painter.setPen(QPen(QColor(255, 240, 80), 2))
                painter.drawLine(x, peak_y, x + bar_w - 1, peak_y)

            if clipped_count > 0:
                painter.fillRect(x, 0, bar_w, 5, QColor(255, 40, 40))


# ─── Filter toolbar helpers ──────────────────────────────────────────────────

def _make_combo(parent, items):
    """Create a QComboBox from a list of (data_key, display_label) tuples."""
    from classes.app import get_app
    _ = get_app()._tr
    cb = QComboBox(parent)
    cb.setSizeAdjustPolicy(QComboBox.AdjustToContents)
    for key, label in items:
        cb.addItem(_(label), key)
    return cb


def _restore_combo(combo, value):
    for i in range(combo.count()):
        if combo.itemData(i) == value:
            combo.setCurrentIndex(i)
            return


# ─── Waveform dock content (painter + toolbar) ───────────────────────────────

class WaveformDockContent(QWidget):
    """Waveform dock widget: filter toolbar above the waveform painter."""
    scopeRegionToggled = pyqtSignal(bool)
    renderSettingsChanged = pyqtSignal()

    _MODES = [
        ("luma",        N_("Luma")),
        ("rgb_overlay", N_("RGB Overlay")),
        ("rgb_parade",  N_("RGB Parade")),
        ("red",         N_("Red")),
        ("green",       N_("Green")),
        ("blue",        N_("Blue")),
    ]
    _COLORS = [
        ("green",  N_("Green")),
        ("white",  N_("White")),
        ("orange", N_("Orange")),
    ]
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFocusPolicy(Qt.NoFocus)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(8, 4, 8, 6)
        layout.setSpacing(4)

        toolbar = QWidget(self)
        toolbar.setFocusPolicy(Qt.NoFocus)
        tl = QHBoxLayout(toolbar)
        tl.setContentsMargins(0, 0, 0, 0)
        tl.setSpacing(6)

        self._mode_cb  = _make_combo(toolbar, self._MODES)
        self._color_cb = _make_combo(toolbar, self._COLORS)
        self._region_btn = _make_scope_region_button(toolbar)
        tl.addWidget(self._mode_cb)
        tl.addWidget(self._color_cb)
        tl.addStretch()
        tl.addWidget(self._region_btn)

        self.waveform = WaveformWidget(self)
        layout.addWidget(toolbar)
        layout.addWidget(self.waveform)

        # Restore saved state
        _restore_combo(self._mode_cb,  _get(_S_WAVE_MODE,  "luma"))
        _restore_combo(self._color_cb, _get(_S_WAVE_COLOR, "green"))
        self._sync_color_visibility()

        self._mode_cb.currentIndexChanged.connect(self._on_mode)
        self._color_cb.currentIndexChanged.connect(self._on_color)
        self._region_btn.toggled.connect(self.scopeRegionToggled.emit)

    def _sync_color_visibility(self):
        self._color_cb.setVisible(self._mode_cb.currentData() == "luma")

    def _on_mode(self):
        self.waveform.set_mode(self._mode_cb.currentData())
        self._sync_color_visibility()
        self.renderSettingsChanged.emit()

    def _on_color(self):
        self.waveform.set_color(self._color_cb.currentData())

    def render_settings(self):
        return {
            "columns": max(32, int(self.waveform.width() or 256)),
        }

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.renderSettingsChanged.emit()

    def set_scope_region_enabled(self, enabled):
        self._region_btn.blockSignals(True)
        self._region_btn.setChecked(bool(enabled))
        self._region_btn.blockSignals(False)

    @pyqtSlot(dict)
    def update_data(self, video_data):
        self.waveform.update_data(video_data)


# ─── Histogram dock content (painter + toolbar) ──────────────────────────────

class HistogramDockContent(QWidget):
    """Histogram dock widget: filter toolbar above the histogram painter."""
    scopeRegionToggled = pyqtSignal(bool)

    _CHANNELS = [
        ("rgba",  N_("All Channels")),
        ("luma",  N_("Luma")),
        ("red",   N_("Red")),
        ("green", N_("Green")),
        ("blue",  N_("Blue")),
    ]
    _SCALES = [
        ("log",    N_("Logarithmic")),
        ("linear", N_("Linear")),
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFocusPolicy(Qt.NoFocus)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(8, 4, 8, 6)
        layout.setSpacing(4)

        toolbar = QWidget(self)
        toolbar.setFocusPolicy(Qt.NoFocus)
        tl = QHBoxLayout(toolbar)
        tl.setContentsMargins(0, 0, 0, 0)
        tl.setSpacing(6)

        self._ch_cb    = _make_combo(toolbar, self._CHANNELS)
        self._scale_cb = _make_combo(toolbar, self._SCALES)
        self._region_btn = _make_scope_region_button(toolbar)
        tl.addWidget(self._ch_cb)
        tl.addWidget(self._scale_cb)
        tl.addStretch()
        tl.addWidget(self._region_btn)

        self.histogram = HistogramWidget(self)
        layout.addWidget(toolbar)
        layout.addWidget(self.histogram)

        _restore_combo(self._ch_cb,    _get(_S_HIST_CH,    "rgba"))
        _restore_combo(self._scale_cb, _get(_S_HIST_SCALE, "log"))

        self._ch_cb.currentIndexChanged.connect(self._on_channel)
        self._scale_cb.currentIndexChanged.connect(self._on_scale)
        self._region_btn.toggled.connect(self.scopeRegionToggled.emit)

    def _on_channel(self):
        self.histogram.set_channel(self._ch_cb.currentData())

    def _on_scale(self):
        self.histogram.set_scale(self._scale_cb.currentData())

    def set_scope_region_enabled(self, enabled):
        self._region_btn.blockSignals(True)
        self._region_btn.setChecked(bool(enabled))
        self._region_btn.blockSignals(False)

    @pyqtSlot(dict)
    def update_data(self, video_data):
        self.histogram.update_data(video_data)


class VectorscopeDockContent(QWidget):
    """Vectorscope dock widget with a minimal toolbar."""
    scopeRegionToggled = pyqtSignal(bool)
    renderSettingsChanged = pyqtSignal()

    _DISPLAYS = [
        ("colorized", N_("Colorized")),
        ("density", N_("Density")),
        ("intensity", N_("Intensity")),
    ]
    _ZOOMS = [
        ("100", "100%"),
        ("200", "200%"),
        ("400", "400%"),
    ]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFocusPolicy(Qt.NoFocus)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(8, 4, 8, 6)
        layout.setSpacing(4)

        toolbar = QWidget(self)
        toolbar.setFocusPolicy(Qt.NoFocus)
        tl = QHBoxLayout(toolbar)
        tl.setContentsMargins(0, 0, 0, 0)
        tl.setSpacing(6)

        self._display_cb = _make_combo(toolbar, self._DISPLAYS)
        self._zoom_cb = _make_combo(toolbar, self._ZOOMS)
        self._region_btn = _make_scope_region_button(toolbar)
        tl.addWidget(self._display_cb)
        tl.addWidget(self._zoom_cb)
        tl.addStretch()
        tl.addWidget(self._region_btn)

        self.vectorscope = VectorscopeWidget(self)
        layout.addWidget(toolbar)
        layout.addWidget(self.vectorscope)

        _restore_combo(self._display_cb, _get(_S_VEC_DISPLAY, "colorized"))
        _restore_combo(self._zoom_cb, _get(_S_VEC_ZOOM, "100"))

        self._display_cb.currentIndexChanged.connect(self._on_display)
        self._zoom_cb.currentIndexChanged.connect(self._on_zoom)
        self._region_btn.toggled.connect(self.scopeRegionToggled.emit)

    def _on_display(self):
        self.vectorscope.set_display(self._display_cb.currentData())
        self.renderSettingsChanged.emit()

    def _on_zoom(self):
        self.vectorscope.set_zoom(self._zoom_cb.currentData())
        self.renderSettingsChanged.emit()

    def render_settings(self):
        return {
            "display": normalize_vectorscope_display(self._display_cb.currentData()),
            "zoom": str(self._zoom_cb.currentData() or "100"),
        }

    def set_scope_region_enabled(self, enabled):
        self._region_btn.blockSignals(True)
        self._region_btn.setChecked(bool(enabled))
        self._region_btn.blockSignals(False)

    @pyqtSlot(dict)
    def update_data(self, video_data):
        self.vectorscope.update_data(video_data)
